from dataclasses import dataclass, field

import sqlglot


@dataclass(kw_only=True)
class SqlViewGenerator:
    """A class to generate SQL view statements given a select
    statement or subquery.

    Generated SQL is compatible with mysql, postgres and sqlite3.

    The given `subquery` is not validated.

    For use with view definitions.
    """

    report_model: str = None
    ordering: list[str] | None = field(default_factory=list)
    order_by: str | None = field(init=False, default=None)
    with_stmt: str | None = field(default="")
    footer: str = field(init=False, default=None)
    template: str | None = field(
        default="{with_stmt} {select_stmt} from ({subquery}) as A ORDER BY {order_by};"
    )

    def __post_init__(self):
        ordering = [f"{fld[1:]} desc" if fld.startswith("~") else fld for fld in self.ordering]
        self.order_by = ", ".join(ordering) or "subject_identifier"
        self.footer = f") as A ORDER BY {self.order_by}"

    @staticmethod
    def transpile(sql: str, read: str | None = None, write: str | None = None) -> str:
        read = read or "mysql"
        sql = sql.replace(";", "")
        return sqlglot.transpile(sql, read=read, write=write)[0]

    def as_mysql(self, subquery: str, read: str | None = None) -> str:
        select_stmt = f"select *, uuid() as id, now() as `created`, '{self.report_model}' as `report_model`"  # noqa
        subquery = self.transpile(subquery, read=read, write="mysql")
        return self.template.format(
            with_stmt=self.with_stmt,
            select_stmt=select_stmt,
            subquery=subquery,
            order_by=self.order_by,
        )

    def as_postgres(self, subquery: str, read: str | None = None) -> str:
        select_stmt = f"select *, get_random_uuid() as id, now() as created, '{self.report_model}' as report_model"  # noqa
        subquery = self.transpile(subquery, read=read, write="postgres")
        return self.template.format(
            with_stmt=self.with_stmt,
            select_stmt=select_stmt,
            subquery=subquery,
            order_by=self.order_by,
        )

    def as_sqlite(self, subquery: str, read: str | None = None) -> str:
        """For UUID support in sqlite, install sqlite/uuid.
        See https://github.com/nalgeon/sqlpkg-cli?tab=readme-ov-file
        and https://sqlpkg.org/?q=uuid

            $ curl -sS https://webi.sh/sqlpkg | sh
            $ sqlpkg install sqlite/uuid
        """
        select_stmt = f"select *, uuid() as id, datetime() as created, '{self.report_model}' as report_model"  # noqa
        subquery = self.transpile(subquery, read=read, write="sqlite")
        return self.template.format(
            with_stmt=self.with_stmt,
            select_stmt=select_stmt,
            subquery=subquery,
            order_by=self.order_by,
        )



from .research_protocol_config import ResearchProtocolConfig


class ResearchProtocolConfigMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    def process_template_response(self, request, response):  # noqa: ARG002
        if getattr(response, "context_data", None):
            protocol_config = ResearchProtocolConfig()
            response.context_data.update(
                copyright=protocol_config.copyright,
                disclaimer=protocol_config.disclaimer,
                institution=protocol_config.institution,
                license=protocol_config.license,
                project_name=protocol_config.project_name,
            )
        return response



from __future__ import annotations

from django.forms import ModelForm

from edc_model_form.mixins.report_datetime_modelform_mixin import (
    ReportDatetimeModelFormMixin,
)


class FormValidatorMixin:
    """A ModelForm mixin to add a validator class.

    Declare with `forms.ModelForm`. If with multiple mixins,
    declare last before ModelForm:

        class MyForm(mixin1, mixin2, FormValidatorMixin, ModelForm):
            ...

    """

    form_validator_cls = None

    def clean(self: ModelForm) -> dict:
        cleaned_data = super().clean()
        try:
            form_validator = self.form_validator_cls(
                cleaned_data=cleaned_data,
                instance=self.instance,
                data=self.data,
                model=self._meta.model,
                current_site=getattr(self, "current_site", None),
                locale=getattr(self, "current_locale", None),
            )
        except TypeError as e:
            if str(e) != "'NoneType' object is not callable":
                raise
        else:
            cleaned_data = form_validator.validate()
        return cleaned_data


class ReportDatetimeFormValidatorMixin(ReportDatetimeModelFormMixin):
    """to be declared with FormValidators."""

    report_datetime_field_attr: str = "report_datetime"



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING
from uuid import UUID

from django.contrib.sites.models import Site

from edc_view_utils import ModelButton

from ..models import ActionType

if TYPE_CHECKING:
    from edc_action_item.action import Action
    from edc_action_item.models import ActionItem
    from edc_appointment.models import Appointment


@dataclass(kw_only=True)
class ActionItemButton(ModelButton):
    action_cls: type[Action] = None
    model_obj: ActionItem | None = None
    model_cls: type[ActionItem] = field(default=None)
    appointment: Appointment = None
    next_url_name: str = field(default="subject_dashboard_url")

    @property
    def label(self) -> str | None:
        if self.fixed_label:
            return self.fixed_label
        return self.action_cls.reference_model_cls()._meta.verbose_name

    @property
    def site(self) -> Site | None:
        """If model_obj is None, then Site should come from the
        request object.
        """
        return (
            getattr(self.model_obj, "site", None)
            or getattr(self.appointment, "site", None)
            or getattr(self.request, "site", None)
        )

    def get_subject_identifier(self):
        if self.appointment:
            return self.appointment.subject_identifier
        return self.subject_identifier

    @property
    def url(self) -> str:
        url = super().url
        return url

    @property
    def reverse_kwargs(self) -> dict[str, str | UUID]:
        if self.appointment:
            return dict(
                subject_identifier=self.appointment.subject_identifier,
                appointment=self.appointment.id,
            )
        return dict(
            subject_identifier=self.subject_identifier,
        )

    @property
    def extra_kwargs(self) -> dict[str, UUID]:
        return dict(
            action_type=ActionType.objects.get(name=self.action_cls.name).id,
            priority=self.action_cls.priority,
            subject_identifier=self.get_subject_identifier(),
            appointment=getattr(self.appointment, "id", None),
            related_action_item_id=getattr(self.model_obj, "related_action_item_id", ""),
            parent_action_item_id=getattr(self.model_obj, "parent_action_item_id", ""),
        )



# Generated by Django 2.0.7 on 2018-07-17 12:42

import _socket
import django_audit_fields.fields.uuid_auto_field
import django_revision.revision_field
from django.db import migrations, models

import edc_model.validators.date
import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.other_charfield
import edc_model_fields.fields.userfield
import edc_protocol.validators
import django.utils.timezone


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="SubjectOffstudy",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "subject_identifier",
                    models.CharField(
                        max_length=50, unique=True, verbose_name="Subject Identifier"
                    ),
                ),
                (
                    "offstudy_datetime",
                    models.DateTimeField(
                        default=django.utils.timezone.now,
                        validators=[
                            edc_protocol.validators.datetime_not_before_study_start,
                            edc_model.validators.date.datetime_not_future,
                        ],
                        verbose_name="Off-study date and time",
                    ),
                ),
                (
                    "offstudy_reason",
                    models.CharField(
                        choices=[
                            ("lost", "Lost to follow-up"),
                            ("completed protocol", "Completed protocol"),
                            ("consent_withdrawal", "Completed protocol"),
                            ("dead", "Deceased"),
                        ],
                        max_length=125,
                        verbose_name="Please code the primary reason participant taken off-study",
                    ),
                ),
                (
                    "offstudy_reason_other",
                    edc_model_fields.fields.other_charfield.OtherCharField(
                        blank=True,
                        max_length=35,
                        null=True,
                        verbose_name="If Other, specify ...",
                    ),
                ),
            ],
            options={"abstract": False},
        )
    ]



# Generated by Django 4.2.7 on 2023-12-03 00:22

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_lab", "0028_alter_aliquot_device_created_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="aliquot",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Aliquot",
            },
        ),
        migrations.AlterModelOptions(
            name="box",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "ordering": ("-box_datetime",),
                "verbose_name": "Box",
                "verbose_name_plural": "Boxes",
            },
        ),
        migrations.AlterModelOptions(
            name="boxitem",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "ordering": ("position",),
                "verbose_name": "Box Item",
            },
        ),
        migrations.AlterModelOptions(
            name="boxtype",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "ordering": ("name",),
                "verbose_name": "Box Type",
            },
        ),
        migrations.AlterModelOptions(
            name="consignee",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "ordering": ("name",),
                "verbose_name": "Consignee",
            },
        ),
        migrations.AlterModelOptions(
            name="manifest",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "ordering": ("-manifest_identifier",),
                "verbose_name": "Manifest",
            },
        ),
        migrations.AlterModelOptions(
            name="manifestitem",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "ordering": ("created",),
                "verbose_name": "Manifest Item",
            },
        ),
        migrations.AlterModelOptions(
            name="order",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Order",
            },
        ),
        migrations.AlterModelOptions(
            name="panel",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Panel",
                "verbose_name_plural": "Panels",
            },
        ),
        migrations.AlterModelOptions(
            name="result",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Result",
            },
        ),
        migrations.AlterModelOptions(
            name="resultitem",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Result Item",
            },
        ),
        migrations.AlterModelOptions(
            name="shipper",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "ordering": ("name",),
                "verbose_name": "Shipper",
            },
        ),
        migrations.AddField(
            model_name="aliquot",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="aliquot",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="box",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="box",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="boxitem",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="boxitem",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="boxtype",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="boxtype",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="consignee",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="consignee",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalaliquot",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalaliquot",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalbox",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalbox",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalboxitem",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalboxitem",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalconsignee",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalconsignee",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalmanifest",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalmanifest",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalorder",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalorder",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalresult",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalresult",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalresultitem",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalresultitem",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="historicalshipper",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="historicalshipper",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="manifest",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="manifest",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="manifestitem",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="manifestitem",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="order",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="order",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="panel",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="panel",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="result",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="result",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="resultitem",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="resultitem",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddField(
            model_name="shipper",
            name="locale_created",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale created",
            ),
        ),
        migrations.AddField(
            model_name="shipper",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                null=True,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AddIndex(
            model_name="aliquot",
            index=models.Index(
                fields=["-modified", "-created"], name="edc_lab_ali_modifie_49b6da_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="aliquot",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_lab_ali_user_mo_5b1f7d_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="box",
            index=models.Index(
                fields=["-modified", "-created"], name="edc_lab_box_modifie_af5133_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="box",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_lab_box_user_mo_2bc680_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="boxitem",
            index=models.Index(
                fields=["-modified", "-created"], name="edc_lab_box_modifie_6a2a97_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="boxitem",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_lab_box_user_mo_63be01_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="boxtype",
            index=models.Index(
                fields=["-modified", "-created"], name="edc_lab_box_modifie_eb0207_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="boxtype",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_lab_box_user_mo_b65019_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="consignee",
            index=models.Index(fields=["name"], name="edc_lab_con_name_2de87e_idx"),
        ),
        migrations.AddIndex(
            model_name="manifest",
            index=models.Index(
                fields=["-modified", "-created"], name="edc_lab_man_modifie_aae866_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="manifest",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_lab_man_user_mo_14c069_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="manifestitem",
            index=models.Index(
                fields=["-modified", "-created"], name="edc_lab_man_modifie_113dc0_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="manifestitem",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_lab_man_user_mo_198958_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="order",
            index=models.Index(
                fields=["-modified", "-created"], name="edc_lab_ord_modifie_33b427_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="order",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_lab_ord_user_mo_990f68_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="panel",
            index=models.Index(
                fields=["lab_profile_name", "name"],
                name="edc_lab_pan_lab_pro_2a0c6d_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="result",
            index=models.Index(
                fields=["-modified", "-created"], name="edc_lab_res_modifie_edced5_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="result",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_lab_res_user_mo_49e9bc_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="resultitem",
            index=models.Index(
                fields=["-modified", "-created"], name="edc_lab_res_modifie_d7258d_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="resultitem",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_lab_res_user_mo_437ddc_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="shipper",
            index=models.Index(
                fields=["-modified", "-created"], name="edc_lab_shi_modifie_822b3b_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="shipper",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_lab_shi_user_mo_2a93c7_idx",
            ),
        ),
    ]



from decimal import Decimal

from django.core.validators import MinValueValidator
from django.db import models
from django.utils import timezone
from edc_model.models import BaseUuidModel, HistoricalRecords
from sequences import get_next_value

from ...exceptions import RepackRequestError
from .container import Container


class Manager(models.Manager):
    use_in_migrations = True


class RepackRequest(BaseUuidModel):
    """A model to repack stock from one container into another.

    Move stock from one phycical container into another, for example
    move stock from a bottle of 50000 into x number of containers
    of 128.

    Location is not changed here.
    """

    repack_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    repack_datetime = models.DateTimeField(default=timezone.now)

    from_stock = models.ForeignKey(
        "edc_pharmacy.stock",
        on_delete=models.PROTECT,
        related_name="repack_requests",
        null=True,
        blank=False,
        limit_choices_to={"repack_request__isnull": True},
    )

    container = models.ForeignKey(
        Container,
        on_delete=models.PROTECT,
        null=True,
        blank=False,
        limit_choices_to={"may_repack_as": True},
    )

    container_unit_qty = models.DecimalField(
        verbose_name="Container unit quantity",
        decimal_places=2,
        max_digits=10,
        null=True,
        blank=True,
        validators=[MinValueValidator(Decimal("1.0"))],
        help_text="Leave blank for default.",
    )

    override_container_unit_qty = models.BooleanField(default=False)

    item_qty_repack = models.IntegerField(
        verbose_name="Number of containers to repack",
        null=True,
        blank=False,
        validators=[MinValueValidator(0)],
    )

    item_qty_processed = models.IntegerField(
        verbose_name="Number of containers processed", blank=False, default=0
    )

    unit_qty_processed = models.DecimalField(
        verbose_name="Unit quantity processed",
        default=Decimal("0.0"),
        blank=False,
        decimal_places=2,
        max_digits=10,
        help_text="Automatically calculated",
    )

    stock_count = models.IntegerField(null=True, blank=True)

    task_id = models.UUIDField(null=True)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return self.repack_identifier

    def save(self, *args, **kwargs):
        if not self.repack_identifier:
            next_id = get_next_value(self._meta.label_lower)
            self.repack_identifier = f"{next_id:06d}"
            self.processed = False
        if not self.from_stock.confirmed:
            raise RepackRequestError(
                "Unconfirmed stock item. Only confirmed stock items may "
                "be used to repack. Perhaps catch this in the form"
            )
        self.container_unit_qty = self.container_unit_qty or self.container.unit_qty_default
        self.item_qty_processed = (
            0 if self.item_qty_processed is None else self.item_qty_processed
        )
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Repack request"
        verbose_name_plural = "Repack request"



import sys

from django.core.management.color import color_style

style = color_style()


def update_action_types(sender=None, verbose=None, **kwargs):
    from .site_action_items import site_action_items

    sys.stdout.write(style.MIGRATE_HEADING("Updating action types:\n"))
    site_action_items.create_or_update_action_types()
    sys.stdout.write("Done.\n")
    sys.stdout.flush()



# Generated by Django 5.2.10 on 2026-01-24 20:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0117_alter_historicalorderitem_container_unit_qty_and_more"),
    ]

    operations = [
        migrations.RenameField(
            model_name="container",
            old_name="max_per_subject",
            new_name="max_items_per_subject",
        ),
        migrations.RenameField(
            model_name="container",
            old_name="qty_default",
            new_name="unit_qty_default",
        ),
        migrations.RenameField(
            model_name="container",
            old_name="qty_places",
            new_name="unit_qty_places",
        ),
        migrations.RenameField(
            model_name="historicalcontainer",
            old_name="max_per_subject",
            new_name="max_items_per_subject",
        ),
        migrations.RenameField(
            model_name="historicalcontainer",
            old_name="qty_default",
            new_name="unit_qty_default",
        ),
        migrations.RenameField(
            model_name="historicalcontainer",
            old_name="qty_places",
            new_name="unit_qty_places",
        ),
        migrations.RemoveField(
            model_name="container",
            name="qty_may_be_adjusted_from_default",
        ),
        migrations.RemoveField(
            model_name="historicalcontainer",
            name="qty_may_be_adjusted_from_default",
        ),
        migrations.AddField(
            model_name="container",
            name="unit_qty_max",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Maximum unit capacity of container",
                max_digits=10,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="historicalcontainer",
            name="unit_qty_max",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Maximum unit capacity of container",
                max_digits=10,
                null=True,
            ),
        ),
    ]



from __future__ import annotations

from clinicedc_constants import (
    FUTURE_DATE,
    IS_NULL,
    LAST_MONTH,
    LAST_WEEK,
    NOT_NULL,
    PAST_DATE,
    THIS_MONTH,
    THIS_WEEK,
    TODAY,
    YESTERDAY,
)
from dateutil.relativedelta import MO, relativedelta
from django.contrib.admin import SimpleListFilter
from django.db.models import QuerySet
from django.utils import timezone
from django.utils.translation import gettext as _


class PastDateListFilter(SimpleListFilter):
    title = None

    parameter_name = None
    field_name = None

    def lookups(self, request, model_admin) -> tuple:
        return (
            (TODAY, _("Today")),
            (YESTERDAY, _("Yesterday")),
            (THIS_WEEK, _("This week")),
            (LAST_WEEK, _("Last week")),
            (THIS_MONTH, _("This month")),
            (LAST_MONTH, _("Last month")),
            (PAST_DATE, _("Any past date")),
            (FUTURE_DATE, _("Any future date")),
            (IS_NULL, _("No date")),
            (NOT_NULL, _("Has date")),
        )

    @property
    def extra_queryset_options(self) -> dict:
        return {}

    def queryset(self, request, queryset) -> QuerySet | None:
        morning = timezone.now().replace(second=0, hour=0, minute=0)
        monday = morning + relativedelta(weekday=MO(-1))
        night = timezone.now().replace(second=59, hour=23, minute=59)
        qs = None
        if self.value() == THIS_WEEK:
            qs = queryset.filter(
                **{
                    f"{self.field_name}__gte": monday,
                    f"{self.field_name}__lt": monday + relativedelta(weeks=1),
                },
                **self.extra_queryset_options,
            ).order_by(self.field_name)
        elif self.value() == TODAY:
            qs = queryset.filter(
                **{
                    f"{self.field_name}__gte": morning,
                    f"{self.field_name}__lt": night,
                },
                **self.extra_queryset_options,
            ).order_by(self.field_name)
        elif self.value() == YESTERDAY:
            qs = queryset.filter(
                **{
                    f"{self.field_name}__lte": night,
                    f"{self.field_name}__lt": night - relativedelta(days=1),
                },
                **self.extra_queryset_options,
            ).order_by(self.field_name)
        elif self.value() == LAST_WEEK:
            qs = queryset.filter(
                **{
                    f"{self.field_name}__gte": monday,
                    f"{self.field_name}__lt": monday - relativedelta(weeks=1),
                },
                **self.extra_queryset_options,
            ).order_by(f"-{self.field_name}")
        elif self.value() == THIS_MONTH:
            qs = queryset.filter(
                **{
                    f"{self.field_name}__gte": monday,
                    f"{self.field_name}__lt": monday + relativedelta(months=1),
                },
                **self.extra_queryset_options,
            ).order_by(f"-{self.field_name}")
        elif self.value() == LAST_MONTH:
            qs = queryset.filter(
                **{
                    f"{self.field_name}__gte": monday,
                    f"{self.field_name}__lt": monday - relativedelta(months=1),
                },
                **self.extra_queryset_options,
            ).order_by(f"-{self.field_name}")
        elif self.value() == PAST_DATE:
            qs = queryset.filter(
                **{f"{self.field_name}__lt": morning},
                **self.extra_queryset_options,
            ).order_by(f"-{self.field_name}")
        elif self.value() == FUTURE_DATE:
            qs = queryset.filter(
                **{f"{self.field_name}__gt": night},
                **self.extra_queryset_options,
            ).order_by(self.field_name)
        elif self.value() == NOT_NULL:
            qs = queryset.filter(
                **{f"{self.field_name}__isnull": False},
                **self.extra_queryset_options,
            ).order_by(self.field_name)
        elif self.value() == IS_NULL:
            qs = queryset.filter(
                **{f"{self.field_name}__isnull": True},
                **self.extra_queryset_options,
            ).order_by(self.field_name)
        return qs



from django.db import models
from django.db.models.deletion import PROTECT

from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_sites.managers import CurrentSiteManager

from ..model_mixins import ResultItemModelMixin
from .result import Result


class ResultItemManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, report_datetime, requisition_identifier):
        return self.get(
            report_datetime=report_datetime,
            requisition__requisition_identifier=requisition_identifier,
        )


class ResultItem(ResultItemModelMixin, BaseUuidModel):
    result = models.ForeignKey(Result, on_delete=PROTECT)

    objects = ResultItemManager()

    on_site = CurrentSiteManager()

    history = HistoricalRecords()

    def natural_key(self):
        return self.report_datetime, *self.result.natural_key()

    natural_key.dependencies = ("edc_lab.result", "sites.Site")

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Result Item"



# declared here to avoid importing edc-action-item
# may need to check to ensure this is the current
# list of fields
ACTION_ITEM_COLUMNS = [
    "action_identifier",
    "action_item_id",
    "action_item_reason",
    "parent_action_identifier",
    "parent_action_item_id",
    "related_action_identifier",
    "related_action_item_id",
]

# declared here to avoid importing edc-model
# may need to check to ensure this is the current
# list of fields
SYSTEM_COLUMNS = [
    "created",
    "modified",
    "user_created",
    "user_modified",
    "hostname_created",
    "hostname_modified",
    "device_created",
    "device_modified",
    "locale_created",
    "locale_modified",
    "revision",
]



# Generated by Django 5.2.3 on 2025-07-09 03:09

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0086_auto_20250708_2337"),
    ]

    operations = [
        migrations.AlterField(
            model_name="storagebinitem",
            name="stock",
            field=models.OneToOneField(
                limit_choices_to={
                    "confirmationatsiteitem__isnull": False,
                    "dispenseitem__isnull": True,
                },
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.stock",
            ),
        ),
    ]



from __future__ import annotations

from edc_crf.crf_form_validator import CrfFormValidator

from .requisition_form_validator_mixin import RequisitionFormValidatorMixin


class RequisitionFormValidator(RequisitionFormValidatorMixin, CrfFormValidator):
    """Form validator for requisitions (e.g. SubjectRequisition)"""

    report_datetime_field_attr: str = "requisition_datetime"



"""
Here only so we can run makemessages

"""

from datetime import datetime
from pathlib import Path
from zoneinfo import ZoneInfo

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/5.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-+#-*m&byi1w8=#e5nr))^kollb=k1b#rvjffo#+teji^qpm14o"  # nosec B105  # noqa: S105

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []

DJANGO_CRYPTO_FIELDS_KEY_PATH = BASE_DIR / ".etc"
AUTO_CREATE_KEYS = True
APP_NAME = "edc_pharmacy"

SUBJECT_VISIT_MODEL = "edc_visit_tracking.subjectvisit"
EDC_PROTOCOL_STUDY_OPEN_DATETIME = datetime(2013, 10, 15, tzinfo=ZoneInfo("Africa/Gaborone"))
EDC_PROTOCOL_STUDY_CLOSE_DATETIME = datetime(2033, 10, 15, tzinfo=ZoneInfo("Africa/Gaborone"))
EDC_ADVERSE_EVENT_APP_LABEL = "edc_adverse_event"
# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.sites",
    "django_crypto_fields.apps.AppConfig",
    "django_pylabels.apps.AppConfig",
    "edc_pylabels.apps.AppConfig",
    "edc_sites.apps.AppConfig",
    "edc_notification.apps.AppConfig",
    "edc_action_item.apps.AppConfig",
    "edc_registration.apps.AppConfig",
    "edc_visit_schedule.apps.AppConfig",
    "edc_visit_tracking.apps.AppConfig",
    "edc_appointment.apps.AppConfig",
    "edc_pharmacy.apps.AppConfig",
]


MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "edc_model_admin.context_processors.admin_theme",
                "edc_constants.context_processor.constants",
                "edc_appointment.context_processors.constants",
                "edc_visit_tracking.context_processors.constants",
            ]
        },
    }
]

ROOT_URLCONF = "edc_pharmacy.urls"


WSGI_APPLICATION = "edc_pharmacy.wsgi.application"


# Database
# https://docs.djangoproject.com/en/5.1/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    }
}


# Password validation
# https://docs.djangoproject.com/en/5.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/5.1/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/5.1/howto/static-files/

STATIC_URL = "static/"

# Default primary key field type
# https://docs.djangoproject.com/en/5.1/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"



from .baseline_review_model_mixin_factory import baseline_review_model_mixin_factory
from .dx_initial_review_model_mixin_factory import dx_initial_review_model_mixin_factory
from .followup_review_model_mixin_factory import followup_review_model_mixin_factory
from .rx_initial_review_model_mixin_factory import rx_initial_review_model_mixin_factory



import sys

from django.apps import apps as django_apps
from django.core.management.color import color_style

style = color_style()


def post_migrate_user_groups_and_roles(sender=None, **kwargs):
    """Update Groups, Role model with EDC defaults."""

    from .auth_updater import AuthUpdater  # noqa: PLC0415
    from .auth_updater.group_updater import CodenameDoesNotExist  # noqa: PLC0415

    try:
        AuthUpdater(apps=django_apps, verbose=True)
    except CodenameDoesNotExist as e:
        sys.stdout.write(style.ERROR(f"{e}. "))
        sys.stdout.write(
            style.ERROR(
                "\n\nIf the codename is mispelled, correct the error and try again.\n\n"
            )
        )
        sys.stdout.write(
            style.ERROR(
                "\nIf the codename is a custom codename (not Django's), "
                "make sure you created it.\n\n"
            )
        )
        sys.stdout.write(
            style.ERROR(
                "\nIf this is happening in a migration that is creating a new model,\n"
                "the post_migrate signal that creates the new model's permissions\n"
                "might be queued to run AFTER edc_auth's post_migrate signal. Let this\n"
                "migration complete and then run `migrate` again.\n\n"
                "If the problem persists, check your code.\n\n"
            )
        )



# Generated by Django 5.2.10 on 2026-01-25 01:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0120_auto_20260125_0103"),
    ]

    operations = [
        migrations.AddField(
            model_name="confirmation",
            name="code",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                max_length=15,
                verbose_name="Stock code",
            ),
        ),
        migrations.AddField(
            model_name="historicalconfirmation",
            name="code",
            field=models.CharField(
                blank=True,
                default="",
                editable=False,
                max_length=15,
                verbose_name="Stock code",
            ),
        ),
        migrations.AlterField(
            model_name="container",
            name="display_name",
            field=models.CharField(default="", max_length=100, unique=True),
        ),
        migrations.AlterField(
            model_name="container",
            name="name",
            field=models.CharField(max_length=50, unique=True),
        ),
        migrations.AlterField(
            model_name="historicalcontainer",
            name="display_name",
            field=models.CharField(db_index=True, default="", max_length=100),
        ),
        migrations.AlterField(
            model_name="historicalcontainer",
            name="name",
            field=models.CharField(db_index=True, max_length=50),
        ),
        migrations.AlterField(
            model_name="historicalreceiveitem",
            name="container_unit_qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Number of units in the container(s). Must be the same for all containers in the record.",
                max_digits=10,
                null=True,
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreceiveitem",
            name="reference",
            field=models.CharField(default="", max_length=150),
        ),
        migrations.AlterField(
            model_name="historicalreceiveitem",
            name="unit_qty_received",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="number of containers x container unit qty, e.g. 10 x Bottle of 128 = 1280. Quantiy will recalculate after save.",
                max_digits=10,
                null=True,
                verbose_name="Unit quantity received",
            ),
        ),
        migrations.AlterField(
            model_name="receiveitem",
            name="container_unit_qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Number of units in the container(s). Must be the same for all containers in the record.",
                max_digits=10,
                null=True,
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AlterField(
            model_name="receiveitem",
            name="reference",
            field=models.CharField(default="", max_length=150),
        ),
        migrations.AlterField(
            model_name="receiveitem",
            name="unit_qty_received",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="number of containers x container unit qty, e.g. 10 x Bottle of 128 = 1280. Quantiy will recalculate after save.",
                max_digits=10,
                null=True,
                verbose_name="Unit quantity received",
            ),
        ),
    ]



from pathlib import Path
from urllib.error import URLError
from urllib.parse import urljoin
from urllib.request import urlretrieve

from django.conf import settings


def get_static_file(app_label: str, filename: str) -> str:
    path = Path(settings.STATIC_ROOT or "") / app_label / filename
    # path as os path
    if path.is_file():
        try:
            with path.open("r"):
                pass
        except FileNotFoundError:
            path = None
    else:
        path = None

    # path as a url
    if not path:
        path = urljoin(f"https://{settings.STATIC_URL}", app_label, filename)
        try:
            urlretrieve(path)  # noqa: S310
        except URLError:
            path = None
    if not path:
        raise FileNotFoundError(
            f"Static file not found. Tried "
            f"STATIC_ROOT ({settings.STATIC_ROOT}) and "
            f"STATIC_URL ({settings.STATIC_URL}). "
            f"Got {app_label}/{filename}."
        )
    return path



from edc_model_admin.admin_site import EdcAdminSite

from .apps import AppConfig

edc_pylabels_admin = EdcAdminSite(
    name="edc_pylabels_admin",
    app_label=AppConfig.name,
    keep_delete_action=True,
)



from .base_model_form_mixin import BaseModelFormMixin, BaseModelFormMixinError
from .inline_model_form_mixin import InlineModelFormMixin, InlineModelFormMixinError

__all__ = [
    "BaseModelFormMixin",
    "BaseModelFormMixinError",
    "InlineModelFormMixin",
    "InlineModelFormMixinError",
]



from edc_auth.get_app_codenames import get_app_codenames

EDC_FACILITY = "EDC_FACILITY"
EDC_FACILITY_VIEW = "EDC_FACILITY_VIEW"
EDC_FACILITY_SUPER = "EDC_FACILITY_SUPER"

codenames = get_app_codenames("edc_facility")
codenames = [
    codename
    for codename in codenames
    if codename
    not in [
        "edc_facility.add_healthfacilitytypes",
        "edc_facility.change_healthfacilitytypes",
        "edc_facility.delete_healthfacilitytypes",
        "edc_facility.add_holiday",
        "edc_facility.change_holiday",
        "edc_facility.delete_holiday",
    ]
]



from __future__ import annotations

from django import forms
from django.contrib import admin

from ..admin_site import edc_metadata_admin
from ..models import CrfMetadata
from .modeladmin_mixins import MetadataModelAdminMixin


class CrfMetadataForm(forms.ModelForm):
    class Meta:
        model = CrfMetadata
        fields = "__all__"
        verbose_name = "CRF collection status"


@admin.register(CrfMetadata, site=edc_metadata_admin)
class CrfMetadataAdmin(MetadataModelAdminMixin):
    form = CrfMetadataForm
    ordering = ()
    changelist_url = "edc_metadata_admin:edc_metadata_crfmetadata_changelist"
    change_list_title = "CRF collection status"
    change_form_title = "CRF collection status"



from edc_navbar import Navbar, NavbarItem, site_navbars

label = Navbar(name="edc_label")

label.register(
    NavbarItem(
        name="label",
        label="Label",
        fa_icon="fa-film",
        url_with_namespace="edc_label:home_url",
        codename="edc_navbar.nav_label",
    )
)

site_navbars.register(label)



from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.translation import gettext as _
from django_audit_fields import audit_fieldset_tuple
from rangefilter.filters import DateRangeFilterBuilder

from edc_model_admin.history import SimpleHistoryAdmin
from edc_utils.date import to_local

from ...admin_site import edc_pharmacy_admin
from ...models import ConfirmationAtLocation
from ..model_admin_mixin import ModelAdminMixin


@admin.register(ConfirmationAtLocation, site=edc_pharmacy_admin)
class ConfirmationAtLocationAdmin(ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Stock confirmations at location"
    change_form_title = "Pharmacy: Stock confirmation at location"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    ordering = ("-transfer_confirmation_identifier",)

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "transfer_confirmation_identifier",
                    "transfer_confirmation_datetime",
                    "stock_transfer",
                    "location",
                    "comments",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "identifier",
        "transfer_confirmation_date",
        "location",
        "confirmation_at_location_item_changelist",
        "stock_transfer_changelist",
    )

    list_filter = (
        "location",
        ("transfer_confirmation_datetime", DateRangeFilterBuilder()),
    )

    readonly_fields = (
        "transfer_confirmation_identifier",
        "transfer_confirmation_datetime",
        "stock_transfer",
        "location",
    )

    search_fields = (
        "pk",
        "transfer_confirmation_identifier",
        "stock_transfer__pk",
        "confirmationatlocationitem__code",
        (
            "confirmationatlocationitem__stock_transfer_item__stock__allocation__registered_subject__subject_identifier"
        ),
    )

    @admin.display(description="CONFIRMATION #", ordering="-transfer_confirmation_identifier")
    def identifier(self, obj):
        return obj.transfer_confirmation_identifier.split("-")[0]

    @admin.display(description="Date", ordering="transfer_confirmation_datetime")
    def transfer_confirmation_date(self, obj):
        return to_local(obj.transfer_confirmation_datetime).date()

    @admin.display(description="Confirmed items")
    def confirmation_at_location_item_changelist(self, obj):
        item_count = obj.confirmationatlocationitem_set.all().count()
        url = reverse("edc_pharmacy_admin:edc_pharmacy_confirmationatlocationitem_changelist")
        url = f"{url}?q={obj.id}"
        context = dict(
            url=url,
            label=_("Confirmed items (%(item_count)s)") % {"item_count": item_count},
            title="Go to stock transfer confirmation items",
        )
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(
        description="Stock Transfer", ordering="stock_transfer__transfer_identifier"
    )
    def stock_transfer_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stocktransfer_changelist")
        url = f"{url}?q={obj.stock_transfer.id}"
        context = dict(
            url=url,
            label=(
                f"{obj.stock_transfer.transfer_identifier} ({obj.stock_transfer.item_count})"
            ),
            title="Go to stock transfer",
        )
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)



from clinicedc_constants import OTHER

list_data = {
    "edc_transfer.transferreasons": [
        ("moved", "Moved away from the area"),
        ("unhappy_with_care", "Was unhappy with the care they are receiving"),
        ("need_specialized_treatment", "Needed more specialist treatment"),
        ("drug_supply", "Due to drug supply "),
        ("stigma", "Concerned about stigma"),
        (OTHER, "Other reason (specify below)"),
    ],
}



# Generated by Django 5.2.10 on 2026-01-25 02:19

import django.core.validators
from decimal import Decimal
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0122_auto_20260125_0402"),
    ]

    operations = [
        migrations.RenameField(
            model_name="historicalrepackrequest",
            old_name="requested_qty",
            new_name="repack_item_qty",
        ),
        migrations.RenameField(
            model_name="repackrequest",
            old_name="requested_qty",
            new_name="repack_item_qty",
        ),
        migrations.AddField(
            model_name="historicalrepackrequest",
            name="container_unit_qty",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="Leave blank for default. Number of units in the container(s). Must be the same for all containers in the record.",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AddField(
            model_name="repackrequest",
            name="container_unit_qty",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="Leave blank for default. Number of units in the container(s). Must be the same for all containers in the record.",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AlterField(
            model_name="container",
            name="max_items_per_subject",
            field=models.IntegerField(
                blank=True,
                help_text="Maximum number of this container that may be allocated to a subject per stock request. (For example, no more than 3 bottles per subject)",
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
            ),
        ),
        migrations.AlterField(
            model_name="container",
            name="unit_qty_default",
            field=models.DecimalField(
                decimal_places=2,
                help_text="May be adjusted. Default value for entry form",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
                verbose_name="Default qty",
            ),
        ),
        migrations.AlterField(
            model_name="container",
            name="unit_qty_max",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Maximum unit capacity of container",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
            ),
        ),
        migrations.AlterField(
            model_name="container",
            name="unit_qty_places",
            field=models.IntegerField(
                default=0.0,
                help_text="May NOT be adjusted at receiving",
                validators=[django.core.validators.MinValueValidator(0)],
            ),
        ),
        migrations.AlterField(
            model_name="historicalcontainer",
            name="max_items_per_subject",
            field=models.IntegerField(
                blank=True,
                help_text="Maximum number of this container that may be allocated to a subject per stock request. (For example, no more than 3 bottles per subject)",
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
            ),
        ),
        migrations.AlterField(
            model_name="historicalcontainer",
            name="unit_qty_default",
            field=models.DecimalField(
                decimal_places=2,
                help_text="May be adjusted. Default value for entry form",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
                verbose_name="Default qty",
            ),
        ),
        migrations.AlterField(
            model_name="historicalcontainer",
            name="unit_qty_max",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Maximum unit capacity of container",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
            ),
        ),
        migrations.AlterField(
            model_name="historicalcontainer",
            name="unit_qty_places",
            field=models.IntegerField(
                default=0.0,
                help_text="May NOT be adjusted at receiving",
                validators=[django.core.validators.MinValueValidator(0)],
            ),
        ),
        migrations.AlterField(
            model_name="historicalorder",
            name="item_count",
            field=models.IntegerField(
                null=True,
                validators=[django.core.validators.MinValueValidator(1)],
                verbose_name="Item count",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorderitem",
            name="container_unit_qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Number of units in the container(s). Must be the same for all containers in the record.",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorderitem",
            name="item_qty_ordered",
            field=models.IntegerField(
                help_text="Number of containers",
                null=True,
                validators=[django.core.validators.MinValueValidator(1)],
                verbose_name="Number of containers ordered",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorderitem",
            name="unit_qty_ordered",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Updated automatically (containers * unit_quantity per container)",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
                verbose_name="Unit quantity ordered",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorderitem",
            name="unit_qty_pending",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Unit quantity ordered - Unit quantity received. Updated automatically",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("0.0"))],
                verbose_name="Unit quantity pending",
            ),
        ),
        migrations.AlterField(
            model_name="historicalorderitem",
            name="unit_qty_received",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Updated automatically when units are received",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("0.0"))],
                verbose_name="Unit quantity received",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreceive",
            name="item_count",
            field=models.IntegerField(
                null=True,
                validators=[django.core.validators.MinValueValidator(1)],
                verbose_name="Item count",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreceiveitem",
            name="container_unit_qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Number of units in the container(s). Must be the same for all containers in the record.",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreceiveitem",
            name="item_qty_received",
            field=models.IntegerField(
                help_text="Partial containers should be received separately",
                null=True,
                validators=[django.core.validators.MinValueValidator(1)],
                verbose_name="Number of containers received",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreceiveitem",
            name="unit_qty_received",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="number of containers x container unit qty, e.g. 10 x Bottle of 128 = 1280. Quantiy will recalculate after save.",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
                verbose_name="Unit quantity received",
            ),
        ),
        migrations.AlterField(
            model_name="historicalstock",
            name="dispensed",
            field=models.BooleanField(default=False, help_text="See dispenseitem."),
        ),
        migrations.AlterField(
            model_name="order",
            name="item_count",
            field=models.IntegerField(
                null=True,
                validators=[django.core.validators.MinValueValidator(1)],
                verbose_name="Item count",
            ),
        ),
        migrations.AlterField(
            model_name="orderitem",
            name="container_unit_qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Number of units in the container(s). Must be the same for all containers in the record.",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AlterField(
            model_name="orderitem",
            name="item_qty_ordered",
            field=models.IntegerField(
                help_text="Number of containers",
                null=True,
                validators=[django.core.validators.MinValueValidator(1)],
                verbose_name="Number of containers ordered",
            ),
        ),
        migrations.AlterField(
            model_name="orderitem",
            name="unit_qty_ordered",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Updated automatically (containers * unit_quantity per container)",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
                verbose_name="Unit quantity ordered",
            ),
        ),
        migrations.AlterField(
            model_name="orderitem",
            name="unit_qty_pending",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Unit quantity ordered - Unit quantity received. Updated automatically",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("0.0"))],
                verbose_name="Unit quantity pending",
            ),
        ),
        migrations.AlterField(
            model_name="orderitem",
            name="unit_qty_received",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Updated automatically when units are received",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("0.0"))],
                verbose_name="Unit quantity received",
            ),
        ),
        migrations.AlterField(
            model_name="receive",
            name="item_count",
            field=models.IntegerField(
                null=True,
                validators=[django.core.validators.MinValueValidator(1)],
                verbose_name="Item count",
            ),
        ),
        migrations.AlterField(
            model_name="receiveitem",
            name="container_unit_qty",
            field=models.DecimalField(
                decimal_places=2,
                help_text="Number of units in the container(s). Must be the same for all containers in the record.",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
                verbose_name="Container unit quantity",
            ),
        ),
        migrations.AlterField(
            model_name="receiveitem",
            name="item_qty_received",
            field=models.IntegerField(
                help_text="Partial containers should be received separately",
                null=True,
                validators=[django.core.validators.MinValueValidator(1)],
                verbose_name="Number of containers received",
            ),
        ),
        migrations.AlterField(
            model_name="receiveitem",
            name="unit_qty_received",
            field=models.DecimalField(
                blank=True,
                decimal_places=2,
                help_text="number of containers x container unit qty, e.g. 10 x Bottle of 128 = 1280. Quantiy will recalculate after save.",
                max_digits=10,
                null=True,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
                verbose_name="Unit quantity received",
            ),
        ),
        migrations.AlterField(
            model_name="stock",
            name="dispensed",
            field=models.BooleanField(default=False, help_text="See dispenseitem."),
        ),
    ]



# Generated by Django 3.2.13 on 2022-09-29 14:42

import django.db.models.manager
from django.db import migrations

import edc_sites.models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_identifier", "0004_auto_20200513_0042"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="identifiermodel",
            managers=[
                ("objects", django.db.models.manager.Manager()),
                ("on_site", edc_sites.models.CurrentSiteManager()),
            ],
        ),
    ]



from django.contrib import admin

from edc_list_data.admin import ListModelAdminMixin

from .admin_site import edc_facility_admin
from .modeladmin_mixins import HealthFacilityModelAdminMixin
from .models import HealthFacility, HealthFacilityTypes, Holiday


@admin.register(Holiday, site=edc_facility_admin)
class HolidayAdmin(admin.ModelAdmin):
    date_hierarchy = "local_date"
    list_display = ("name", "local_date")


@admin.register(HealthFacility, site=edc_facility_admin)
class HealthFacilityAdmin(HealthFacilityModelAdminMixin, admin.ModelAdmin):
    pass


@admin.register(HealthFacilityTypes, site=edc_facility_admin)
class HealthFacilityTypesAdmin(ListModelAdminMixin, admin.ModelAdmin):
    pass



from __future__ import annotations

from typing import TYPE_CHECKING

from edc_protocol.research_protocol_config import ResearchProtocolConfig
from reportlab.graphics.barcode.widgets import BarcodeCode128
from reportlab.graphics.shapes import Drawing, String

from ..utils import format_qty
from .draw_label_watermark import draw_label_watermark

if TYPE_CHECKING:
    from ..models import Stock


def draw_patient_stock_label_code128(
    label: Drawing,
    width: int | float,
    height: int | float,
    obj: Stock,
) -> Drawing:
    """Callable to draw a single study medication label given a model
    instance `obj`
    """

    draw_label_watermark(label, width, height)

    br = BarcodeCode128(humanReadable=True, barHeight=30, barWidth=0.7, gap=1.7)
    br.value = obj.code
    br.x = 0
    br.y = height - 40
    label.add(br)

    qty = format_qty(obj.container_unit_qty, obj.container)
    formulation = obj.product.formulation
    product = (
        f"{formulation.medication} {int(formulation.strength)}"
        f"{formulation.get_units_display()} "
    )
    label.add(String(15, height - 72, ResearchProtocolConfig().protocol_name, fontSize=10))
    label.add(String(15, height - 84, f"{product}", fontSize=10))
    label.add(String(15, height - 96, f"{qty} tabs", fontSize=10))
    return label



from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from django_audit_fields.admin import audit_fieldset_tuple

from edc_model_admin.history import SimpleHistoryAdmin
from edc_utils.date import to_local

from ...admin_site import edc_pharmacy_admin
from ...forms import ReceiveForm
from ...models import Receive
from ..actions import confirm_received_stock_action, print_labels_from_receive
from ..model_admin_mixin import ModelAdminMixin


@admin.register(Receive, site=edc_pharmacy_admin)
class ReceiveAdmin(ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Receiving"
    change_form_title = "Pharmacy: Receive"
    history_list_display = ()

    show_object_tools = False
    show_cancel = True
    list_per_page = 20

    form = ReceiveForm
    ordering = ("-receive_identifier",)
    actions = (print_labels_from_receive, confirm_received_stock_action)
    autocomplete_fields = ("supplier",)

    fieldsets = (
        (
            None,
            {"fields": ("receive_identifier",)},
        ),
        (
            "Section A",
            {
                "fields": (
                    "receive_datetime",
                    "order",
                    "location",
                )
            },
        ),
        (
            "Section B",
            {
                "fields": (
                    "supplier",
                    "invoice_number",
                    "invoice_date",
                )
            },
        ),
        ("Section C: Comment / Notes", {"fields": ("comment",)}),
        audit_fieldset_tuple,
    )

    list_display = (
        "identifier",
        "receive_date",
        "location",
        "order_changelist",
        "items",
        "stock_changelist",
        "supplier",
        "invoice_number",
        "invoice_date",
        "created",
        "modified",
    )
    list_filter = (
        "receive_datetime",
        "location",
        "invoice_date",
        "created",
        "modified",
    )
    search_fields = (
        "id",
        "receive_identifier",
        "order__id",
        "location__name",
        "invoice_number",
        "supplier__name",
    )

    @admin.display(description="RECEIVE #", ordering="receive_identifier")
    def identifier(self, obj):
        return obj.receive_identifier

    @admin.display(description="Receive date", ordering="receive_datetime")
    def receive_date(self, obj):
        return to_local(obj.receive_datetime).date()

    @admin.display(description="Received items", ordering="receive_identifier")
    def items(self, obj):
        count = obj.receiveitem_set.all().count()
        url = reverse("edc_pharmacy_admin:edc_pharmacy_receiveitem_changelist")
        url = f"{url}?q={obj.id}"
        context = dict(url=url, label=f"Received ({count})", title="Go to received items")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Order #")
    def order_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_order_changelist")
        url = f"{url}?q={obj.order.order_identifier!s}"
        context = dict(url=url, label=obj.order.order_identifier, title="Back to order")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Stock")
    def stock_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stock_changelist")
        url = f"{url}?q={obj.id}"
        context = dict(url=url, label="Stock", title="Go to stock")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "order" and request.GET.get("order"):
            kwargs["queryset"] = db_field.related_model.objects.filter(
                pk=request.GET.get("order", 0)
            )
        return super().formfield_for_foreignkey(db_field, request, **kwargs)



from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from edc_model.admin import HistoricalModelAdminMixin
from edc_model_admin.history import SimpleHistoryAdmin
from edc_utils import to_local
from rangefilter.filters import DateRangeFilterBuilder

from ....admin_site import edc_pharmacy_history_admin
from ....models import ConfirmationAtLocationItem
from ...model_admin_mixin import ModelAdminMixin


@admin.register(ConfirmationAtLocationItem.history.model, site=edc_pharmacy_history_admin)
class ConfirmationAtLocationItemHistoryAdmin(
    ModelAdminMixin, HistoricalModelAdminMixin, SimpleHistoryAdmin
):
    change_list_title = "Pharmacy: Stock confirmation items at location (History)"
    change_form_title = "Pharmacy: Stock confirmation item at location (History)"
    change_list_note_url_name = (
        "edc_pharmacy_admin:edc_pharmacy_confirmationatlocationitem_changelist"
    )
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    list_display = (
        "item_id",
        "history_type",
        "formatted_history_date",
        "location_name",
        "code",
        "confirm_at_location_changelist",
        "revision",
        "history_id",
    )

    list_filter = (
        "history_type",
        ("history_date", DateRangeFilterBuilder()),
        ("transfer_confirmation_item_datetime", DateRangeFilterBuilder()),
    )

    search_fields = (
        "code",
        "history_id",
        "transfer_confirmation_item_identifier",
        "confirm_at_location__transfer_confirmation_identifier",
    )

    @admin.display(description="ITEM#", ordering="transfer_confirmation_item_identifier")
    def item_id(self, obj):
        return obj.transfer_confirmation_item_identifier

    @admin.display(description="Date", ordering="transfer_confirmation_item_datetime")
    def item_date(self, obj):
        if obj.transfer_confirmation_item_datetime:
            return to_local(obj.transfer_confirmation_item_datetime).date()
        return None

    @admin.display(
        description="Location", ordering="confirmation_at_location__location__display_name"
    )
    def location(self, obj):
        if obj:
            return obj.confirmation_at_location.location.display_name
        return None

    @admin.display(description="Confirmation#")
    def confirm_at_location_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_confirmationatlocation_changelist")
        url = f"{url}?q={obj.confirm_at_location.transfer_confirmation_identifier}"
        context = dict(
            url=url,
            label=f"{obj.confirm_at_location.transfer_confirmation_identifier}",
            title="Go to confirmation at site",
        )
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)



from __future__ import annotations

from decimal import Decimal

from django.db import models
from django.utils import timezone
from edc_list_data.model_mixins import BaseListModelMixin, ListModelMixin
from edc_model.models import BaseUuidModel

from .stock import Stock


class StockTransactionType(ListModelMixin):
    class Meta(BaseListModelMixin.Meta):
        verbose_name = "Stock transaction type"
        verbose_name_plural = "Stock transaction types"


class StockTransaction(BaseUuidModel):
    stock = models.ForeignKey(Stock, on_delete=models.PROTECT)

    transaction_datetime = models.DateTimeField(default=timezone.now)

    transaction_type = models.ForeignKey(StockTransactionType, on_delete=models.PROTECT)

    transaction_qty = models.DecimalField(
        blank=False,
        decimal_places=2,
        max_digits=20,
        default=Decimal("0.0"),
    )

    balance = models.DecimalField(
        blank=False,
        decimal_places=2,
        max_digits=20,
        default=Decimal("0.0"),
    )

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Stock transaction"
        verbose_name_plural = "Stock transactions"



# Generated by Django 5.2.3 on 2025-07-02 21:13

from decimal import Decimal

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0077_historicalstockadjustment_adjustment_datetime_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalstock",
            name="qty",
            field=models.DecimalField(
                decimal_places=2,
                default=Decimal("0"),
                help_text="Difference of qty_in and qty_out",
                max_digits=20,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalstock",
            name="qty_in",
            field=models.DecimalField(
                decimal_places=2,
                default=Decimal("0"),
                help_text="Container qty, e.g. 1 bucket, 1 bottle, etc",
                max_digits=20,
                null=True,
                validators=[
                    django.core.validators.MinValueValidator(0),
                    django.core.validators.MaxValueValidator(1),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="historicalstock",
            name="qty_out",
            field=models.DecimalField(
                decimal_places=2,
                default=Decimal("0"),
                help_text="Container qty, e.g. 1 bucket, 1 bottle, etc",
                max_digits=20,
                validators=[
                    django.core.validators.MinValueValidator(0),
                    django.core.validators.MaxValueValidator(1),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="historicalstock",
            name="unit_qty_in",
            field=models.DecimalField(
                decimal_places=2,
                default=Decimal("0"),
                help_text="Number of units in this container, e.g. 128 tablets. See post-save signal",
                max_digits=20,
                validators=[django.core.validators.MinValueValidator(0)],
            ),
        ),
        migrations.AlterField(
            model_name="historicalstock",
            name="unit_qty_out",
            field=models.DecimalField(
                decimal_places=2,
                default=Decimal("0"),
                help_text="Number of units from this container, e.g. 128 tablets",
                max_digits=20,
                validators=[django.core.validators.MinValueValidator(0)],
            ),
        ),
        migrations.AlterField(
            model_name="stock",
            name="qty",
            field=models.DecimalField(
                decimal_places=2,
                default=Decimal("0"),
                help_text="Difference of qty_in and qty_out",
                max_digits=20,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="stock",
            name="qty_in",
            field=models.DecimalField(
                decimal_places=2,
                default=Decimal("0"),
                help_text="Container qty, e.g. 1 bucket, 1 bottle, etc",
                max_digits=20,
                null=True,
                validators=[
                    django.core.validators.MinValueValidator(0),
                    django.core.validators.MaxValueValidator(1),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="stock",
            name="qty_out",
            field=models.DecimalField(
                decimal_places=2,
                default=Decimal("0"),
                help_text="Container qty, e.g. 1 bucket, 1 bottle, etc",
                max_digits=20,
                validators=[
                    django.core.validators.MinValueValidator(0),
                    django.core.validators.MaxValueValidator(1),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="stock",
            name="unit_qty_in",
            field=models.DecimalField(
                decimal_places=2,
                default=Decimal("0"),
                help_text="Number of units in this container, e.g. 128 tablets. See post-save signal",
                max_digits=20,
                validators=[django.core.validators.MinValueValidator(0)],
            ),
        ),
        migrations.AlterField(
            model_name="stock",
            name="unit_qty_out",
            field=models.DecimalField(
                decimal_places=2,
                default=Decimal("0"),
                help_text="Number of units from this container, e.g. 128 tablets",
                max_digits=20,
                validators=[django.core.validators.MinValueValidator(0)],
            ),
        ),
    ]



# Generated by Django 2.1.4 on 2019-01-14 00:50

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_locator", "0016_auto_20181108_0353")]

    operations = [
        migrations.AlterField(
            model_name="historicalsubjectlocator",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.Site",
            ),
        )
    ]



# Generated by Django 4.2.7 on 2023-12-04 22:19

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("edc_dx_review", "0003_alter_diagnosislocations_options_and_more"),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name="diagnosislocations",
            name="edc_dx_revi_name_a39b40_idx",
        ),
        migrations.RemoveIndex(
            model_name="reasonsfortesting",
            name="edc_dx_revi_name_0b2844_idx",
        ),
    ]



# Generated by Django 4.2.7 on 2023-12-04 22:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_metadata", "0027_alter_crfmetadata_options_and_more"),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name="requisitionmetadata",
            name="edc_metadat_subject_51288a_idx",
        ),
        migrations.RemoveIndex(
            model_name="requisitionmetadata",
            name="edc_metadat_entry_s_865e45_idx",
        ),
        migrations.RenameIndex(
            model_name="crfmetadata",
            new_name="edc_metadat_site_id_d3b213_idx",
            old_name="edc_metadata_crfm_site_ent_idx",
        ),
        migrations.RenameIndex(
            model_name="crfmetadata",
            new_name="edc_metadat_subject_dd4400_idx",
            old_name="edc_metadata_crfm_subjid_v_idx",
        ),
        migrations.AlterUniqueTogether(
            name="requisitionmetadata",
            unique_together=set(),
        ),
        migrations.AddIndex(
            model_name="crfmetadata",
            index=models.Index(
                fields=["subject_identifier"], name="edc_metadat_subject_25f5bc_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="crfmetadata",
            index=models.Index(
                fields=["modified", "created"], name="edc_metadat_modifie_948bce_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="crfmetadata",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_metadat_user_mo_10db17_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="requisitionmetadata",
            index=models.Index(
                fields=["subject_identifier"], name="edc_metadat_subject_feb080_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="requisitionmetadata",
            index=models.Index(
                fields=["modified", "created"], name="edc_metadat_modifie_be0dc2_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="requisitionmetadata",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_metadat_user_mo_a239a3_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="requisitionmetadata",
            index=models.Index(
                fields=[
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_code_sequence",
                    "model",
                    "panel_name",
                    "entry_status",
                    "timepoint",
                    "show_order",
                ],
                name="edc_metadat_subject_04cc77_idx",
            ),
        ),
        migrations.AddConstraint(
            model_name="requisitionmetadata",
            constraint=models.UniqueConstraint(
                fields=(
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_code_sequence",
                    "model",
                    "panel_name",
                ),
                name="edc_metadata_requisitionmetadata_subject_iden_visit_uniq",
            ),
        ),
    ]



from django.apps import apps as django_apps
from django.conf import settings

from .model_mixins import SubjectRefusalModelMixin


def get_subject_refusal_model() -> str:
    return settings.SUBJECT_REFUSAL_MODEL


def get_subject_refusal_model_cls() -> SubjectRefusalModelMixin:
    return django_apps.get_model(get_subject_refusal_model())



# Generated by Django 5.1.2 on 2024-11-15 01:22

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0024_allocation_assignment_containerunits_dispense_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="stock",
            name="receive_item",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.receiveitem",
            ),
        ),
    ]



# Generated by Django 5.0.1 on 2024-02-15 15:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_form_runners", "0002_alter_issue_options_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="issue",
            name="visit_code",
            field=models.CharField(max_length=25, null=True),
        ),
        migrations.AlterField(
            model_name="issue",
            name="visit_code_sequence",
            field=models.IntegerField(null=True),
        ),
    ]



from .address_mixin import AddressMixin
from .base_model import BaseModel
from .base_uuid_model import BaseUuidModel, default_permissions
from .fields import (
    DurationDHField,
    DurationYMDField,
    HostnameModificationField,
    IdentityTypeField,
    InitialsField,
    IsDateEstimatedField,
    IsDateEstimatedFieldNa,
    OtherCharField,
    UserField,
)
from .fields.duration import DurationYearMonthField
from .historical_records import HistoricalRecords
from .name_fields_model_mixin import NameFieldsModelMixin
from .report_status_model_mixin import ReportStatusModelMixin
from .signals import remove_f_expressions
from .url_model_mixin import UrlModelMixin, UrlModelMixinNoReverseMatch



# Generated by Django 5.2.10 on 2026-01-25 23:16
from tqdm import tqdm
from django.db import migrations, transaction

from edc_pharmacy.models import Confirmation, StorageBin


def update_storage_bin_container(apps, schema_editor):
    qs = StorageBin.objects.filter(storagebinitem__isnull=False, container__isnull=True)
    for obj in tqdm(qs, total=qs.count()):
        obj.container = obj.storagebinitem_set.all().first().stock.container
        obj.save(update_fields=["container"])


def update_pharmacy(apps, schema_editor):
    with transaction.atomic():
        qs = Confirmation.objects.all()
        for obj in tqdm(qs, total=qs.count()):
            obj.save()


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0132_historicalstoragebin_container_storagebin_container"),
    ]

    operations = [
        migrations.RunPython(update_pharmacy),
        migrations.RunPython(update_storage_bin_container),
    ]



from django.db import models
from django.db.models import PROTECT

from edc_lab import RequisitionPanel
from edc_lab.utils import get_requisition_model_name
from edc_model.validators import datetime_not_future

requisition_fk_options = dict(
    to=get_requisition_model_name(),
    on_delete=PROTECT,
    related_name="+",
    verbose_name="Requisition",
    null=True,
    blank=True,
    help_text="Start typing the requisition identifier or select one from this visit",
)


class CrfWithRequisitionModelMixin(models.Model):
    """You may also wish to override field `requisition`
    to include `limit_choices_to`.

    For example:
        from edc_lab.model_mixins import (
            CrfWithRequisitionModelMixin,
            requisition_fk_options)

        class MyRequisition(CrfWithRequisitionModelMixin, etc):
            ...
            lab_panel = fbc_panel
            requisition = models.ForeignKey(
                limit_choices_to={"panel__name": fbc_panel.name},
                **requisition_fk_options)
            ...
    """

    lab_panel: RequisitionPanel = None

    requisition = models.ForeignKey(**requisition_fk_options)

    assay_datetime = models.DateTimeField(
        verbose_name="Result assay date and time",
        validators=[datetime_not_future],
        null=True,
        blank=True,
    )

    class Meta:
        abstract = True



from edc_auth.site_auths import site_auths
from edc_auth.utils import remove_default_model_permissions_from_edc_permissions

from .auth_objects import dashboard_tuples


def update_site_auths():
    site_auths.add_post_update_func(
        "edc_dashboard", remove_default_model_permissions_from_edc_permissions
    )

    site_auths.add_custom_permissions_tuples(
        model="edc_dashboard.edcpermissions", codename_tuples=dashboard_tuples
    )


update_site_auths()



import pandas as pd

from ..constants import N_WITH_ROW_PROP, STATISTICS
from ..styler import StylerError


class RowDefinition:
    def __init__(
        self,
        title: str | None = None,
        label: str = None,
        colname: str | None = None,
        condition: pd.Series = None,
        columns: dict[str, tuple[str, int]] = None,
        drop: bool | None = None,
    ):
        """
        A row definition is passed by Table to the RowStatistics
        class.

        :param title:
        :param label:
        :param colname:
        :param condition:
        :param columns: dictionary of label: (statistic, places)=
            {
                FEMALE: (N_WITH_ROW_PROP, 2),
                MALE: (N_WITH_ROW_PROP, 2),
                "All": (N_ONLY, 2),
            }
        :param drop: drops rows from the source dataframe once used by
          the row definition.
        """
        self.title = title or ""
        self.label = label
        self.colname = colname
        self.condition = condition  # condition to filter DF
        self.drop = False if drop is None else drop  # drop index of previous row numerator
        self.columns = columns or {"All": (N_WITH_ROW_PROP, 2)}
        for col, style_info in self.columns.items():
            style, _ = style_info
            if style not in STATISTICS:
                raise StylerError(f"Unknown statistic. Got `{style}` for column `{col}`.")



from edc_lab.models import BoxItem
from edc_listboard.filters import ListboardFilter, ListboardViewFilters


def get_box_items():
    return BoxItem.objects.all().order_by("position").values("identifier")


class RequisitionListboardViewFilters(ListboardViewFilters):
    all = ListboardFilter(name="all", label="All", lookup={})

    received = ListboardFilter(label="Received", lookup={"received": True})

    not_received = ListboardFilter(
        label="Not Received", exclude_filter=True, lookup={"received": True}
    )

    processed = ListboardFilter(label="Processed", lookup={"processed": True})

    not_processed = ListboardFilter(
        label="Not processed", exclude_filter=True, lookup={"processed": True}
    )

    packed = ListboardFilter(label="Packed", lookup={"packed": True})

    not_packed = ListboardFilter(
        label="Not packed", exclude_filter=True, lookup={"packed": True}
    )



# from django.conf import settings
# from django_collect_offline.site_offline_models import site_offline_models
#
# if settings.APP_NAME == 'edc_action_item':
#     exclude_models = []
# else:
#     exclude_models = [
#         'edc_action_item.subjectidentifiermodel',
#         'edc_action_item.TestModelWithoutMixin',
#         'edc_action_item.TestModelWithActionDoesNotCreateAction',
#         'edc_action_item.formzero',
#         'edc_action_item.formone',
#         'edc_action_item.formtwo',
#         'edc_action_item.formthree',
#         'edc_action_item.formfour',
#         'edc_action_item.initial',
#         'edc_action_item.followup',
#         'edc_action_item.appointment',
#         'edc_action_item.subjectvisit',
#         'edc_action_item.crfone',
#         'edc_action_item.crftwo',
#         'edc_action_item.testmodelwithaction']
#
# site_offline_models.register_for_app(
#     'edc_action_item',
#     exclude_models=exclude_models)



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps

if TYPE_CHECKING:
    from ..models import NormalData


__all__ = ["normal_data_model_cls"]


def normal_data_model_cls() -> type[NormalData]:
    return django_apps.get_model("edc_reportable.normaldata")



# Generated by Django 5.1.2 on 2024-11-19 00:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0043_stockproxy_alter_historicallot_lot_no_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="location",
            name="contact_email",
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
        migrations.AddField(
            model_name="location",
            name="contact_name",
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
        migrations.AddField(
            model_name="location",
            name="contact_tel",
            field=models.CharField(blank=True, max_length=150, null=True),
        ),
    ]



from .edc_permissions import EdcPermissions
from .signals import (
    requires_consent_on_pre_save,
    update_appointment_from_consentext_post_save,
)

__all__ = [
    "EdcPermissions",
    "requires_consent_on_pre_save",
    "update_appointment_from_consentext_post_save",
]



from django import forms

from ...models import Location


class LocationForm(forms.ModelForm):

    def clean(self):
        cleaned_data = super().clean()
        if (
            cleaned_data.get("name")
            and cleaned_data.get("site")
            and cleaned_data.get("name") != cleaned_data.get("site").name
        ):
            raise forms.ValidationError({"site": "Site does not match this location"})

        return cleaned_data

    class Meta:
        model = Location
        fields = "__all__"



from django.contrib import admin
from django.contrib.sites.models import Site
from django.template.loader import render_to_string
from django.urls import reverse

from ...admin_site import edc_pharmacy_admin
from ...constants import CENTRAL_LOCATION
from ...models import StorageBinProxy
from .storage_bin_admin import StorageBinAdmin


@admin.register(StorageBinProxy, site=edc_pharmacy_admin)
class StorageBinProxyAdmin(StorageBinAdmin):
    change_list_title = "Pharmacy: Central Storage Bins"
    change_form_title = "Pharmacy: Central Storage Bin"

    @admin.display(description="Items")
    def storage_bin_item_changelist(self, obj):
        items = obj.storagebinitem_set.all().count()
        url = reverse("edc_pharmacy_admin:edc_pharmacy_storagebinitemproxy_changelist")
        url = f"{url}?q={obj.id}"
        context = dict(url=url, label=f"Items ({items})", title="Go to bin items")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    def get_view_only_site_ids_for_user(self, request) -> list[int]:
        return [s.id for s in Site.objects.all()]

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.filter(location__name=CENTRAL_LOCATION)



from __future__ import annotations

from typing import TYPE_CHECKING

from django.forms import ValidationError
from django.test import TestCase

if TYPE_CHECKING:
    from edc_form_validators.form_validator import FormValidator
    from edc_model.models import BaseUuidModel


class FormValidatorTestCaseMixin:
    form_validator_cls: type[FormValidator] = None
    form_validator_model_cls: type[BaseUuidModel] = None

    def validate_form_validator(
        self: FormValidatorTestCaseMixin,
        cleaned_data: dict,
        *,
        instance: BaseUuidModel | None = None,
        model_cls: type[BaseUuidModel] | None = None,
        form_validator_cls: type[FormValidator] | None = None,
    ) -> FormValidator:
        form_validator = (form_validator_cls or self.form_validator_cls)(
            cleaned_data=cleaned_data,
            model=model_cls or self.form_validator_model_cls,
            instance=instance,
        )
        try:
            form_validator.validate()
        except ValidationError:
            pass
        return form_validator

    def assertFormValidatorNoError(  # noqa
        self: FormValidatorTestCaseMixin | TestCase, form_validator
    ) -> None:
        self.assertDictEqual({}, form_validator._errors)

    def assertFormValidatorError(  # noqa
        self: FormValidatorTestCaseMixin | TestCase,
        field: str,
        expected_msg: str,
        form_validator,
        expected_errors: int | None = None,
    ) -> None:
        expected_errors = 1 if expected_errors is None else expected_errors
        self.assertIn(
            field,
            form_validator._errors,
            msg=(
                f"Expected field '{field}' in form validation errors. "
                f"Got '{form_validator._errors}'."
            ),
        )
        self.assertIn(
            expected_msg,
            str(form_validator._errors.get(field)),
            msg=(
                f"Expected error message '{expected_msg}' for field '{field}' "
                f"in form validation errors.  Got '{form_validator._errors}'"
            ),
        )
        self.assertEqual(
            len(form_validator._errors),
            expected_errors,
            msg=(
                f"Expected {expected_errors} error message(s) in form validator. "
                f"Got {len(form_validator._errors)} errors "
                f"as follows: '{form_validator._errors}'"
            ),
        )



from dateutil.relativedelta import relativedelta
from django.utils import timezone

from edc_utils.age import age
from edc_utils.round_up import round_half_away_from_zero

__all__ = ["BMI", "CalculatorError", "calculate_bmi"]


class CalculatorError(Exception):
    pass


MIN_BMI_AGE = 18


class BMI:
    """Calculate BMI, raise if not adult."""

    def __init__(
        self,
        weight_kg=None,
        height_cm=None,
        lower_bmi_value=None,
        upper_bmi_value=None,
        dob=None,
        report_datetime=None,
    ):
        if not weight_kg or not height_cm:
            raise CalculatorError(f"Unable to calculate BMI. Got {weight_kg}kg, {height_cm}cm")
        if age(dob, report_datetime).years < MIN_BMI_AGE:
            raise CalculatorError("Unable to calculate BMI. Got age<18")
        self.lower = float(lower_bmi_value or 5.0)
        self.upper = float(upper_bmi_value or 75.0)
        self.weight = float(weight_kg)
        self.height = float(height_cm) / 100.0
        self.raw_bmi_value = self.weight / (self.height**2)
        if not (self.lower <= self.raw_bmi_value <= self.upper):
            raise CalculatorError(
                "BMI value is absurd. "
                f"Using {self.weight}kg, {self.height}m. Got {self.value}."
            )

    @property
    def value(self):
        return round_half_away_from_zero(self.raw_bmi_value, 4)


def calculate_bmi(
    weight_kg=None,
    height_cm=None,
    lower_bmi_value=None,
    upper_bmi_value=None,
    dob=None,
    report_datetime=None,
    **kwargs,
):
    """Returns a BMI instance or None.

    Assumes adult dob (18) if dob not provided."""
    bmi = None
    if height_cm and weight_kg:
        report_datetime = report_datetime or timezone.now()
        bmi = BMI(
            weight_kg=weight_kg,
            height_cm=height_cm,
            lower_bmi_value=lower_bmi_value,
            upper_bmi_value=upper_bmi_value,
            dob=dob or report_datetime - relativedelta(years=18),
            report_datetime=report_datetime,
        )
    return bmi



from .aliquot_listboard_view_filters import AliquotListboardViewFilters
from .manifest_listboard_filters import ManifestListboardViewFilters
from .pack_listboard_filters import PackListboardViewFilters
from .requisition_listboard_view_filters import RequisitionListboardViewFilters



from clinicedc_constants import MOBILE_NUMBER, NO, OTHER, YES

from .constants import HOSPITAL_NUMBER

IDENTITY_TYPE = (
    ("country_id", "Country ID number"),
    ("drivers", "Driver's license"),
    ("passport", "Passport"),
    (HOSPITAL_NUMBER, "Hospital number"),
    ("country_id_rcpt", "Country ID receipt"),
    (MOBILE_NUMBER, "Mobile number"),
    (OTHER, "Other"),
)

YES_DECLINED = "Yes_declined"

YES_NO_DECLINED_COPY = (
    (YES, "Yes and participant accepted the copy"),
    (NO, "No"),
    (YES_DECLINED, "Yes but participant declined the copy"),
)



from clinicedc_constants import (
    CANCELLED,
    CLOSED,
    HIGH_PRIORITY,
    LOW_PRIORITY,
    MEDIUM_PRIORITY,
    NEW,
    OPEN,
)

ACTION_STATUS = (
    (NEW, "New"),
    (OPEN, "Open"),
    (CLOSED, "Closed"),
    (CANCELLED, "Cancelled"),
)

PRIORITY = (
    (HIGH_PRIORITY, "High Priority"),
    (MEDIUM_PRIORITY, "Medium Priority"),
    (LOW_PRIORITY, "Low Priority"),
)



# Generated by Django 5.2.10 on 2026-01-25 02:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_pharmacy",
            "0125_rename_repack_item_qty_historicalrepackrequest_item_qty_repack_and_more",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalrepackrequest",
            name="item_qty_processed",
            field=models.DecimalField(
                decimal_places=2,
                max_digits=20,
                null=True,
                verbose_name="Number of containers processed",
            ),
        ),
        migrations.AlterField(
            model_name="repackrequest",
            name="item_qty_processed",
            field=models.DecimalField(
                decimal_places=2,
                max_digits=20,
                null=True,
                verbose_name="Number of containers processed",
            ),
        ),
    ]



from __future__ import annotations

from dateutil.relativedelta import relativedelta
from django.contrib.admin import SimpleListFilter
from django.db.models import QuerySet
from django.utils import timezone
from django.utils.translation import gettext as _

from edc_appointment.choices import APPT_STATUS
from edc_appointment.constants import (
    ATTENDED_APPT,
    COMPLETE_APPT,
    GTE_30_TO_60_DAYS,
    GTE_60_TO_90_DAYS,
    GTE_90_TO_180_DAYS,
    GTE_180,
    IN_PROGRESS_APPT,
    INCOMPLETE_APPT,
    LT_30_DAYS,
)
from edc_model_admin.list_filters import FutureDateListFilter


class AppointmentListFilter(FutureDateListFilter):
    title = _("Appointment date")

    parameter_name = "appt_datetime"
    field_name = "appt_datetime"


class AppointmentStatusListFilter(SimpleListFilter):
    title = _("Status")

    parameter_name = "appt_status"
    field_name = "appt_status"

    def lookups(self, request, model_admin) -> tuple[tuple[str, str], ...]:  # noqa: ARG002
        return *APPT_STATUS, (ATTENDED_APPT, "Attended (In progress, incomplete, done)")

    def queryset(self, request, queryset):  # noqa: ARG002
        qs = None
        if self.value() == ATTENDED_APPT:
            qs = queryset.filter(
                appt_status__in=[IN_PROGRESS_APPT, INCOMPLETE_APPT, COMPLETE_APPT]
            )
        elif self.value():
            qs = queryset.filter(appt_status=self.value())
        return qs


class AppointmentOverdueListFilter(SimpleListFilter):
    title = "Overdue (days)"
    parameter_name = "overdue"

    def lookups(self, request, model_admin) -> tuple[tuple[str, str], ...]:  # noqa: ARG002
        return (
            (LT_30_DAYS, _("2-30 days")),
            (GTE_30_TO_60_DAYS, _("30-60 days")),
            (GTE_60_TO_90_DAYS, _("60-90 days")),
            (GTE_90_TO_180_DAYS, _("90-180 days")),
            (GTE_180, _("180+ days")),
        )

    def queryset(self, request, queryset) -> QuerySet | None:  # noqa: ARG002
        now = timezone.now().replace(second=59, hour=23, minute=59)
        qs = None
        if self.value() == LT_30_DAYS:
            qs = queryset.filter(
                appt_datetime__gt=now - relativedelta(days=30),
                appt_datetime__lte=now - relativedelta(days=2),
            ).order_by("appt_datetime")
        elif self.value() == GTE_30_TO_60_DAYS:
            qs = queryset.filter(
                appt_datetime__gt=now - relativedelta(days=60),
                appt_datetime__lte=now - relativedelta(days=30),
            ).order_by("appt_datetime")
        elif self.value() == GTE_60_TO_90_DAYS:
            qs = queryset.filter(
                appt_datetime__gt=now - relativedelta(days=90),
                appt_datetime__lte=now - relativedelta(days=60),
            ).order_by("appt_datetime")
        elif self.value() == GTE_90_TO_180_DAYS:
            qs = queryset.filter(
                appt_datetime__gt=now - relativedelta(days=180),
                appt_datetime__lte=now - relativedelta(days=90),
            ).order_by("appt_datetime")
        elif self.value() == GTE_180:
            qs = queryset.filter(
                appt_datetime__lte=now - relativedelta(days=180),
            ).order_by("appt_datetime")
        return qs



ALLOCATED = "allocated"
ALREADY_CONFIRMED = "already_confirmed"
AVAILABLE = "available"
CANCELLED = "cancelled"
CENTRAL_LOCATION = "central"
CONTROL_ARM = "control_arm"
DAYS = "days"
DISPENSED = "dispensed"
FILLED = "filled"
INVALID = "invalid"
IN_PROGRESS_APPT = "in_progress"
NOT_STORED_AT_LOCATION = "not_stored_at_location"
PARTIAL = "partial"
PER_DAY = "day"
PER_HR = "hr"
PRESCRIPTION_ACTION = "prescription"
PRINT_SELECTED = "print_selected"
PROBLEM = "problem"
RECEIVE = "receive"
REPACK = "repack"
RESERVED = "reserved"
SINGLE = "single"
SOLUTION = "SOLUTION"
ZERO_ITEM = "zero_item"



from edc_registration.models import RegisteredSubject


class QuerySubject(RegisteredSubject):
    class Meta:
        proxy = True
        default_permissions = ("view", "export")
        verbose_name = "Subject"
        verbose_name_plural = "Subjects"



from .consent_definition_form_validator_mixin import ConsentDefinitionFormValidatorMixin
from .subject_consent_form_validator import SubjectConsentFormValidatorMixin

__all__ = ["ConsentDefinitionFormValidatorMixin", "SubjectConsentFormValidatorMixin"]



from django.contrib.admin import AdminSite as DjangoAdminSite
from edc_model_admin.admin_site import EdcAdminSite

from .apps import AppConfig

edc_pharmacy_admin = EdcAdminSite(
    name="edc_pharmacy_admin", app_label=AppConfig.name, keep_delete_action=True
)
edc_pharmacy_admin.disable_action("delete_selected")


class EdcPharmacyHistoryAdmin(DjangoAdminSite):
    site_header = "Pharmacy History"
    site_title = "Pharmacy History"
    index_title = "Pharmacy History"
    index_template = "edc_model_admin/admin/index.html"
    app_index_template = "edc_model_admin/admin/app_index.html"
    login_template = "edc_auth/login.html"
    logout_template = "edc_auth/login.html"
    enable_nav_sidebar = False
    final_catch_all_view = True
    site_url = "/administration/"


edc_pharmacy_history_admin = EdcPharmacyHistoryAdmin(name="edc_pharmacy_history_admin")
edc_pharmacy_history_admin.disable_action("delete_selected")



from pathlib import Path


def get_export_folder() -> Path:
    from django.conf import settings  # noqa: PLC0415

    if path := getattr(settings, "EDC_EXPORT_EXPORT_FOLDER", None):
        return Path(path).expanduser()
    return Path(settings.MEDIA_ROOT) / "data_folder" / "export"



# Generated by Django 5.2.6 on 2025-09-17 16:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_notification", "0010_alter_notification_revision"),
    ]

    operations = [
        migrations.AlterField(
            model_name="notification",
            name="device_created",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device created"
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="device_modified",
            field=models.CharField(
                blank=True, default="", max_length=10, verbose_name="Device modified"
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="locale_created",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale created",
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="locale_modified",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Auto-updated by Modeladmin",
                max_length=10,
                verbose_name="Locale modified",
            ),
        ),
        migrations.AlterField(
            model_name="notification",
            name="mailing_list_address",
            field=models.EmailField(
                blank=True,
                default="",
                max_length=254,
                verbose_name="Mailing list address",
            ),
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist

if TYPE_CHECKING:
    from .models import ActionType


def create_or_update_action_type(
    name: str = None, using: str = None, apps: django_apps = None, **options
) -> ActionType:
    """Returns a model instance of ActionType.

    Gets or creates the ActionType on first pass.

    If model instance exists, updates.
    """

    action_type_model = "edc_action_item.actiontype"
    apps = apps or django_apps

    opts = {}
    action_type_model_cls = apps.get_model(action_type_model)
    if options:
        fields = [f.name for f in action_type_model_cls._meta.get_fields() if f.name != "name"]
        for attr, value in options.items():
            if attr in fields and value is not None:
                opts.update({attr: value})
    try:
        action_type = action_type_model_cls.objects.get(name=name)
    except ObjectDoesNotExist:
        action_type = action_type_model_cls.objects.create(name=name, **opts)
    else:
        if opts:
            for k, v in opts.items():
                setattr(action_type, k, v)
            action_type.save(using=using)
            action_type = action_type_model_cls.objects.get(name=name)
    return action_type



from clinicedc_constants import HIGH_PRIORITY

from edc_action_item.action_with_notification import ActionWithNotification
from edc_offstudy.constants import END_OF_STUDY_ACTION

from .constants import SUBJECT_TRANSFER_ACTION


class SubjectTransferAction(ActionWithNotification):
    reference_model = None  # "inte_prn.subjecttransfer"
    admin_site_name = None  # "inte_prn_admin"

    name = SUBJECT_TRANSFER_ACTION
    display_name = "Submit Subject Transfer"
    notification_display_name = " Subject Transfer"
    parent_action_names = ()
    show_link_to_changelist = True
    show_link_to_add = True
    priority = HIGH_PRIORITY
    singleton = True

    def get_next_actions(self):
        return [END_OF_STUDY_ACTION]



from django.urls import path
from django.views.generic.base import RedirectView

from .admin_site import edc_action_item_admin

app_name = "edc_action_item"

urlpatterns = [
    path("admin/", edc_action_item_admin.urls),
    path("", RedirectView.as_view(url="/edc_action_item/admin/"), name="home_url"),
]



import re
from datetime import date, datetime

from django.core.validators import MaxValueValidator, MinValueValidator, RegexValidator
from django.db import models

from edc_utils import age

from ..exceptions import ValueBoundryError
from ..formula import clean_and_validate_phrase
from .reference_range_collection import ReferenceRangeCollection

MAX_AGE = 130.0


class ReferenceModelMixin(models.Model):
    reference_range_collection = models.ForeignKey(
        ReferenceRangeCollection, on_delete=models.PROTECT
    )

    label = models.CharField(max_length=25)

    description = models.CharField(max_length=255, default="")

    reference_group = models.CharField(max_length=25, default="")

    lower = models.FloatField(null=True)
    lower_operator = models.CharField(max_length=15, default="")
    lower_inclusive = models.BooleanField(default=False)
    lln = models.CharField(max_length=15, default="")

    upper = models.FloatField(null=True)
    upper_operator = models.CharField(max_length=2, default="")
    upper_inclusive = models.BooleanField(default=False)
    uln = models.CharField(max_length=15, default="")

    gender = models.CharField(max_length=1, validators=[RegexValidator(r"[MF]{1}")])
    units = models.CharField(max_length=15)

    age_units = models.CharField(max_length=15)

    age_lower = models.IntegerField()
    age_lower_operator = models.CharField(max_length=15)
    age_lower_inclusive = models.BooleanField(default=False)

    age_upper = models.IntegerField(null=True)
    age_upper_operator = models.CharField(max_length=15, default="")
    age_upper_inclusive = models.BooleanField(null=True)

    fasting = models.BooleanField(default=False)

    phrase = models.CharField(
        max_length=50,
        default="",
        verbose_name="Value phrase",
        help_text="calculated by the formula instance",
    )

    age_phrase = models.CharField(max_length=25, default="", help_text="calculated in save()")

    grade = models.IntegerField(
        null=True, validators=[MinValueValidator(1), MaxValueValidator(5)]
    )

    def save(self, *args, **kwargs):
        self.age_phrase = (
            f"{self.age_lower or ''}{self.age_lower_operator or ''}"
            f"%(age_value)s{self.age_upper_operator or ''}{self.age_upper or ''}"
        )
        self.phrase = self.get_phrase()
        self.description = self.get_description()
        super().save(*args, **kwargs)

    def get_phrase(self) -> str:
        lower = self.lower or ""
        upper = self.upper or ""
        phrase = (
            f"{lower}{self.lln or ''}{self.lower_operator or ''}x"
            f"{self.upper_operator or ''}{upper}{self.uln or ''}"
        )
        clean_and_validate_phrase(phrase)
        return phrase

    def get_description(self, exclude_label: bool | None = None) -> str:
        age_description = (
            ""
            if not self.age_lower and not self.age_upper
            else (
                f"{self.age_lower or ''}{self.age_lower_operator}"
                f"AGE{self.age_upper_operator}{self.age_upper or ''}"
            )
        )
        try:
            fasting = self.fasting
        except KeyError:
            fasting_str = ""
        else:
            fasting_str: str = "Fasting " if fasting else ""
        label = "" if exclude_label else f"{self.label}: "
        grade = ""
        if self.grade is not None:
            grade = f"GRADE{self.grade} "
        return (
            f"{label}"
            f"{self.phrase} "
            f"{self.units} {fasting_str}{grade}{self.gender} "
            f"{age_description}".rstrip()
        )

    def age_in_bounds_or_raise(
        self, dob: date, report_datetime: datetime, age_units: str | None = None
    ) -> bool:
        pattern = r"([<>]=?|==|!=)?\s*-?\d+(\.\d+)?"
        age_units = age_units or "years"
        if age_units not in ["days", "months", "years"]:
            raise ValueError(
                f"Invalid age units. Expected one of {['days', 'months', 'years']}. "
                f"Got {age_units}"
            )
        rdelta = age(dob, report_datetime)
        age_value = getattr(rdelta, age_units)
        if not isinstance(age_value, (int, float)) or not (0.0 <= age_value <= MAX_AGE):
            raise ValueError(f"Invalid age value. Got {age_value}.")
        age_condition_str = self.age_phrase % dict(age_value=age_value)
        if not re.match(pattern, age_condition_str):
            raise ValueError(f"Invalid age condition string. Got {age_condition_str}.")
        if not eval(age_condition_str):  # noqa: S307
            raise ValueBoundryError(
                f"Age is out of bounds. See {self}. Got AGE={age_value} {age_units}."
            )
        return True

    class Meta:
        abstract = True



from __future__ import annotations

from datetime import date, datetime
from typing import TYPE_CHECKING

from clinicedc_utils import convert_units
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.db.models import Q
from django.utils import timezone

from edc_model_to_dataframe.constants import SYSTEM_COLUMNS
from edc_utils import age as get_age

from ..exceptions import NotEvaluated
from .normal_data_model_cls import normal_data_model_cls

if TYPE_CHECKING:
    from django.contrib.sites.models import Site

    from ..models import NormalData, ReferenceRangeCollection


__all__ = ["get_normal_data_or_raise"]


def get_normal_data_or_raise(
    reference_range_collection: ReferenceRangeCollection = None,
    label: str | None = None,
    units: str | None = None,
    gender: str | None = None,
    dob: date | None = None,
    report_datetime: datetime | None = None,
    age_units: str | None = None,
    site: Site | None = None,
    create_missing_normal: bool | None = None,
) -> NormalData:
    # TODO: ensure all age values are inclusive
    obj = None
    age_rdelta = get_age(dob, report_datetime)
    age = getattr(age_rdelta, age_units)
    try:
        obj = normal_data_model_cls().objects.get(
            (Q(age_lower__lte=age) | Q(age_lower__isnull=True)),
            (Q(age_upper__gte=age) | Q(age_upper__isnull=True)),
            reference_range_collection=reference_range_collection,
            label=label,
            gender=gender,
            units=units,
            age_units=age_units,
        )
    except ObjectDoesNotExist as e:
        if create_missing_normal:
            opts = dict(
                reference_range_collection=reference_range_collection,
                label=label,
                gender=gender,
                units=units,
                dob=dob,
                report_datetime=report_datetime,
                age_units=age_units,
            )
            obj = create_obj_for_new_units_or_raise(**opts)
        if not obj:
            raise NotEvaluated(
                f"Value not evaluated. "
                f"Normal reference not found for `{label}`. "
                f"Using units={units}, gender={gender}, "
                f"age={getattr(age_rdelta, age_units)}{age_units}. "
                "Perhaps add this to the default normal reference range data or "
                "pass 'create_missing=True' to convert an existing normal reference."
            ) from e
    except MultipleObjectsReturned as e:
        raise NotEvaluated(
            f"Value not evaluated. "
            f"Multiple normal references found for `{label}`. "
            f"Using units={units}, gender={gender}, age={getattr(age_rdelta, age_units)}. "
        ) from e
    return obj


def create_obj_for_new_units_or_raise(
    reference_range_collection: ReferenceRangeCollection = None,
    label: str | None = None,
    units: str | None = None,
    gender: str | None = None,
    dob: date | None = None,
    report_datetime: datetime | None = None,
    age_units: str | None = None,
) -> NormalData | None:
    opts = {}
    age_rdelta = get_age(dob, report_datetime)
    age = getattr(age_rdelta, age_units)
    # try to find an existing record but with different units
    for obj in (
        normal_data_model_cls()
        .objects.filter(
            (Q(age_lower__lte=age) | Q(age_lower__isnull=True)),
            (Q(age_upper__gte=age) | Q(age_upper__isnull=True)),
            reference_range_collection=reference_range_collection,
            label=label,
            gender=gender,
            age_units=age_units,
        )
        .exclude(units=units)
    ):
        # print(f"Creating normal data: {label} {gender} -- {obj.units}->{units}")
        opts = {
            k: v
            for k, v in obj.__dict__.items()
            if not k.startswith("_")
            and k
            not in [
                "id",
                "units",
                "description",
                "phrase",
                "lower",
                "upper",
                *SYSTEM_COLUMNS,
            ]
        }
        opts["lower"] = convert_units(
            label=obj.label,
            value=obj.lower,
            units_from=obj.units,
            units_to=units,
            places=4,
        )
        opts["upper"] = convert_units(
            label=obj.label,
            value=obj.upper,
            units_from=obj.units,
            units_to=units,
            places=4,
        )
        opts["units"] = units
        opts["auto_created"] = True
        opts["created"] = timezone.now()
        opts["modified"] = opts["created"]
        break
    if opts:
        return normal_data_model_cls().objects.create(**opts)
    return None



from __future__ import annotations

from typing import TYPE_CHECKING

from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured

if TYPE_CHECKING:
    from django.contrib.auth.models import User


def has_profile_or_raise(user: User) -> bool:
    """Raises if user instance does not have a UserProfile
    relation.

    `UserProfile` relation is set up in edc_auth. If `userprofile`
    relation is missing, confirm `edc_auth` is in INSTALLED_APPS.
    """

    user = get_user_model().objects.get(id=user.id)
    userprofile = getattr(user, "userprofile", None)
    if not userprofile:
        raise ImproperlyConfigured(
            "User instance has no `userprofile`. User accounts must have a relation "
            "to `UserProfile`. See edc_sites."
        )
    return True



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone

from .site_visit_schedules import site_visit_schedules

if TYPE_CHECKING:
    from .model_mixins import OnScheduleModelMixin
    from .schedule import Schedule
    from .visit_schedule import VisitSchedule

    class OnScheduleLikeModel(OnScheduleModelMixin): ...


class VisitScheduleViewMixin:
    def __init__(self, **kwargs):
        self.onschedule_models: list[OnScheduleLikeModel] = []
        self.current_schedule: Schedule | None = None
        self.current_visit_schedule: VisitSchedule | None = None
        self.current_onschedule_model: str | None = None
        self.visit_schedules: dict[str, VisitSchedule] = {}
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        # TODO: What if active on more than one schedule??
        for visit_schedule in site_visit_schedules.visit_schedules.values():
            if self.subject_identifier:
                for schedule in visit_schedule.schedules.values():
                    try:
                        onschedule_model_obj = schedule.onschedule_model_cls.objects.get(
                            subject_identifier=self.subject_identifier
                        )
                    except ObjectDoesNotExist:
                        pass
                    else:
                        self.onschedule_models.append(onschedule_model_obj)
                        self.visit_schedules.update(**{visit_schedule.name: visit_schedule})
                        if schedule.is_onschedule(self.subject_identifier, timezone.now()):
                            self.current_schedule = schedule
                            self.current_visit_schedule = visit_schedule
                            self.current_onschedule_model = onschedule_model_obj

        kwargs.update(
            visit_schedules=self.visit_schedules,
            current_onschedule_model=self.current_onschedule_model,
            onschedule_models=self.onschedule_models,
            current_schedule=self.current_schedule,
            current_visit_schedule=self.current_visit_schedule,
        )
        return super().get_context_data(**kwargs)



from __future__ import annotations

from warnings import warn

from django.contrib import messages
from django.contrib.messages import get_messages
from django.core.handlers.wsgi import WSGIRequest
from django.db.models.constants import LOOKUP_SEP
from django.urls.base import reverse
from django.urls.exceptions import NoReverseMatch


class SearchTermLookupError(Exception):
    pass


def get_next_url(request, next_attr=None, warn_to_console=None):
    url = None
    next_url = None
    next_value = request.GET.dict().get(next_attr or "next")
    warn_to_console = True if warn_to_console is None else warn_to_console

    if next_value:
        kwargs = {}
        for pos, value in enumerate(next_value.split(",")):
            if pos == 0:
                next_url = value
            else:
                kwargs.update({value: request.GET.get(value)})
        try:
            url = reverse(next_url, kwargs=kwargs)
        except NoReverseMatch as e:
            if warn_to_console:
                warn(f"{e}. Got {next_value}.")
    return url


def get_value_from_lookup_string(search_field_name: str = None, obj=None, request=None):
    value = None
    for field in search_field_name.split(LOOKUP_SEP):
        if request:
            value = request.GET.get(field, "")
            break
        try:
            value = getattr(value or obj, field)
        except AttributeError as e:
            raise SearchTermLookupError(
                f"Invalid search term. `{search_field_name}`. Got {e}"
            ) from e
        if value is None:
            break
    return value


def add_to_messages_once(
    request: WSGIRequest = None, level: int = None, message: str = None
) -> None:
    if not [msg_obj for msg_obj in get_messages(request) if msg_obj.message == message]:
        messages.add_message(
            request=request,
            message=message,
            level=level or messages.INFO,
            fail_silently=True,
        )



from .historical_note_modeladmin import HistoricalNoteModelAdmin
from .note_modeladmin import NoteModelAdmin
from .qa_report_log_admin import QaReportLogAdmin
from .qa_report_log_summary_admin import QaReportLogSummaryAdmin



# Generated by Django 4.2.11 on 2024-06-19 01:07

import django.db.models.deletion
from django.db import migrations, models

import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("sites", "0002_alter_domain_unique"),
        ("edc_qareports", "0005_edcpermissions"),
    ]

    operations = [
        migrations.CreateModel(
            name="QaReportLog",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("username", models.CharField(max_length=100)),
                ("report_model", models.CharField(max_length=100)),
                ("accessed", models.DateTimeField(default=django.utils.timezone.now)),
                (
                    "site",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="sites.site"
                    ),
                ),
            ],
            options={
                "verbose_name": "QA Report Log",
                "verbose_name_plural": "QA Report Logs",
            },
        ),
    ]



from django.contrib import admin
from django.template.loader import render_to_string
from django.utils.html import format_html
from django.utils.translation import gettext as _
from django_audit_fields.admin import audit_fieldset_tuple


def get_missed_medications_fieldset_tuple():
    description = render_to_string("edc_adherence/missed_medication_fieldset_description.html")
    return (
        _("Missed Medications"),
        {
            "description": format_html("{}", description),
            "fields": (
                "last_missed_pill",
                "missed_pill_reason",
                "other_missed_pill_reason",
            ),
        },
    )


def get_pill_count_fieldset_tuple():
    return (
        _("Pill Count"),
        {
            "fields": (
                "pill_count_performed",
                "pill_count",
                "pill_count_not_performed_reason",
            ),
        },
    )


def get_visual_score_fieldset_tuple(
    description: str | None = None, section_title: str | None = None
):
    section_title = section_title or _("Visual Score")
    description = description or render_to_string(
        "edc_adherence/visual_score_fieldset_description.html"
    )
    return (
        section_title,
        {
            "description": format_html("{}", description),
            "fields": ("visual_score_slider", "visual_score_confirmed"),
        },
    )


class MedicationAdherenceAdminMixin:
    """Declare your admin class using this mixin.

    For example:

        @admin.register(MedicationAdherence, site=my_subject_admin)
        class MedicationAdherenceAdmin(MedicationAdherenceAdminMixin, CrfModelAdmin):

            form = MedicationAdherenceForm

    """

    fieldsets = (
        (None, {"fields": ("subject_visit", "report_datetime")}),
        get_visual_score_fieldset_tuple(),
        get_pill_count_fieldset_tuple(),
        get_missed_medications_fieldset_tuple(),
        audit_fieldset_tuple,
    )

    radio_fields = {  # noqa: RUF012
        "pill_count_performed": admin.VERTICAL,
        "last_missed_pill": admin.VERTICAL,
    }

    filter_horizontal = ("missed_pill_reason",)



class FormulaError(Exception):
    pass


class GradeReferenceError(Exception):
    pass


class LimitsNormalParseError(Exception):
    pass


class ValueReferenceError(Exception):
    pass


class SiteReportablesError(Exception):
    pass


class AlreadyRegistered(Exception):  # noqa: N818
    pass


class ReferenceRangeCollectionError(Exception):
    pass


class ValueBoundryError(Exception):
    pass


class NotEvaluated(Exception):  # noqa: N818
    pass


class BoundariesOverlap(Exception):  # noqa: N818
    pass



# Generated by Django 5.0.7 on 2024-08-12 23:19

import django_db_views.migration_functions
import django_db_views.operations
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_qareports", "0010_auto_20240812_1918"),
    ]

    operations = [
        django_db_views.operations.ViewRunPython(
            code=django_db_views.migration_functions.ForwardViewMigration(
                "select *, uuid() as `id`, now() as `created` from (\n        select site_id, username, report_model, min(accessed) as `first_accessed`,\n               max(accessed) as `last_accessed`, count(*) as `access_count`\n        from edc_qareports_qareportlog\n        group by username, report_model, site_id\n    ) as A\n    order by username, report_model",
                "qa_report_log_summary_view",
                engine="django.db.backends.mysql",
            ),
            reverse_code=django_db_views.migration_functions.BackwardViewMigration(
                "", "qa_report_log_summary_view", engine="django.db.backends.mysql"
            ),
            atomic=False,
        ),
        django_db_views.operations.ViewRunPython(
            code=django_db_views.migration_functions.ForwardViewMigration(
                "select *, get_random_uuid() as id, now() as created (\n        select site_id, username, report_model, min(accessed) as first_accessed,\n               max(accessed) as last_accessed, count(*) as access_count\n        from edc_qareports_qareportlog\n        group by username, report_model, site_id\n    ) as A\n    order by username, report_model",
                "qa_report_log_summary_view",
                engine="django.db.backends.postgresql",
            ),
            reverse_code=django_db_views.migration_functions.BackwardViewMigration(
                "", "qa_report_log_summary_view", engine="django.db.backends.postgresql"
            ),
            atomic=False,
        ),
        django_db_views.operations.ViewRunPython(
            code=django_db_views.migration_functions.ForwardViewMigration(
                "SELECT *, lower(\n    hex(randomblob(4)) || '-' || hex(randomblob(2)) || '-' || '4' ||\n    substr(hex( randomblob(2)), 2) || '-' ||\n    substr('AB89', 1 + (abs(random()) % 4) , 1)  ||\n    substr(hex(randomblob(2)), 2) || '-' ||\n    hex(randomblob(6))\n  ) as id, datetime() as created from (\n        select site_id, username, report_model, min(accessed) as first_accessed,\n               max(accessed) as last_accessed, count(*) as access_count\n        from edc_qareports_qareportlog\n        group by username, report_model, site_id\n    ) as A\n    order by username, report_model",
                "qa_report_log_summary_view",
                engine="django.db.backends.sqlite3",
            ),
            reverse_code=django_db_views.migration_functions.BackwardViewMigration(
                "", "qa_report_log_summary_view", engine="django.db.backends.sqlite3"
            ),
            atomic=False,
        ),
    ]



import sys

from django.conf import settings

from .randomizer import Randomizer
from .site_randomizers import site_randomizers

if getattr(settings, "EDC_RANDOMIZATION_REGISTER_DEFAULT_RANDOMIZER", True):
    site_randomizers.register(Randomizer)
else:
    sys.stdout.write("  -> skipping >>> ")



from django.db import models

from ..models import Notification


class NotificationUserProfileModelMixin(models.Model):
    email_notifications = models.ManyToManyField(
        Notification,
        related_name="email_notifications",
        limit_choices_to={"enabled": True},
        blank=True,
    )

    sms_notifications = models.ManyToManyField(
        Notification,
        related_name="sms_notifications",
        limit_choices_to={"enabled": True},
        blank=True,
    )

    class Meta:
        abstract = True



from django.apps import apps as django_apps
from django.contrib import admin
from django.utils.functional import lazy
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from simple_history.admin import SimpleHistoryAdmin as BaseSimpleHistoryAdmin

format_html_lazy = lazy(format_html, str)


class SimpleHistoryAdmin(BaseSimpleHistoryAdmin):
    history_list_display: tuple[str, ...] = ("dashboard", "change_message")
    object_history_template = "edc_model_admin/admin/object_history.html"
    object_history_form_template = "edc_model_admin/admin/object_history_form.html"

    save_as = False
    save_as_continue = False

    @admin.display(description=_("Change Message"))
    def change_message(self, obj) -> str | None:
        log_entry_model_cls = django_apps.get_model("admin.logentry")
        log_entry = (
            log_entry_model_cls.objects.filter(
                action_time__gte=obj.modified, object_id=str(obj.id)
            )
            .order_by("action_time")
            .first()
        )
        if log_entry:
            return format_html(
                "{html}",
                html=mark_safe(log_entry.get_change_message()),  # nosec B703, B308
            )
        return None

    def dashboard(self, obj) -> str | None:
        if callable(self.view_on_site):
            return format_html_lazy(
                '<A href="{url}">Dashboard</A>',
                url=mark_safe(self.view_on_site()),  # nosec B703, B308
            )
        return None

    def get_list_display(self, request) -> tuple[str, ...]:
        return tuple(super().get_list_display(request))

    def get_list_filter(self, request) -> tuple[str, ...]:
        return tuple(super().get_list_filter(request))

    def get_search_fields(self, request) -> tuple[str, ...]:
        return tuple(super().get_search_fields(request))

    def get_readonly_fields(self, request, obj=None) -> tuple[str, ...]:
        return tuple(super().get_readonly_fields(request, obj=obj))

    def history_view_title(self, request, obj):
        word = _("View") if self.revert_disabled else _("Revert")
        return _("%(word)s %(verbose_name)s Audit Trail") % dict(
            word=str(word), verbose_name=obj._meta.verbose_name.title()
        )

    def history_form_view_title(self, request, obj):
        return self.history_view_title(request, obj)



from clinicedc_constants import HIGH_PRIORITY

from edc_action_item.action_with_notification import ActionWithNotification

from ..constants import HOSPITALIZATION_ACTION


class HospitalizationAction(ActionWithNotification):
    reference_model: str = None  # e.g. "intecomm_prn.hospitalization"
    admin_site_name: str = None  # e.g. "intecomm_prn_admin"

    name = HOSPITALIZATION_ACTION
    display_name = "Submit Hospitalization Report"
    notification_display_name = "Hospitalization"
    parent_action_names = ()
    show_link_to_changelist = True
    show_link_to_add = True
    priority = HIGH_PRIORITY



from django.db import models

from edc_model.models import AddressMixin, BaseUuidModel, HistoricalRecords


class ShipperManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)


class Shipper(AddressMixin, BaseUuidModel):
    name = models.CharField(unique=True, max_length=50)

    objects = ShipperManager()

    history = HistoricalRecords()

    def natural_key(self):
        return (self.name,)

    def __str__(self):
        return self.name

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Shipper"



from .constants import (
    BLOOD_RESULTS_EGFR_ACTION,
    BLOOD_RESULTS_FBC_ACTION,
    BLOOD_RESULTS_GLU_ACTION,
    BLOOD_RESULTS_HBA1C_ACTION,
    BLOOD_RESULTS_INSULIN_ACTION,
    BLOOD_RESULTS_LFT_ACTION,
    BLOOD_RESULTS_LIPIDS_ACTION,
    BLOOD_RESULTS_RFT_ACTION,
)



from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from edc_crf.model_mixins import SingletonCrfModelMixin
from edc_metadata import KEYED
from edc_metadata.utils import refresh_metadata_for_timepoint


@receiver(post_save, weak=False, dispatch_uid="metadata_create_on_post_save")
def metadata_create_on_post_save(
    sender, instance, raw, created, using, update_fields, **kwargs
) -> None:
    """Creates all metadata on post save of model using
    CreatesMetaDataModelMixin.

    For example, when saving the related_visit model.

    A related_visit model instance will:
      * delete and re-create ALL metadata for the timepoint
        (`metadata_create`)
      * run ALL metadata rules for the timepoint
        (`run_metadata_rules_for_related_visit`).
    """
    if (
        not raw
        and not update_fields
        and not hasattr(instance, "metadata_update")
        and not instance._meta.label_lower.split(".")[1].startswith("historical")
    ):
        try:
            instance.metadata_create()
        except AttributeError as e:
            if "metadata_create" not in str(e):
                raise
        else:
            refresh_metadata_for_timepoint(instance, allow_create=True)


@receiver(post_save, weak=False, dispatch_uid="metadata_update_on_post_save")
def metadata_update_on_post_save(
    sender, instance, raw, created, using, update_fields, **kwargs
) -> None:
    """Updates the single metadata record on post save of a
    CRF/Requisition model.

    A CRF/Requisition model instance will:
      * update it`s own references (`update_reference_on_save`)
      * update it`s own metadata (`metadata_update`)
      * run ALL metadata rules for the timepoint
        (`run_metadata_rules_for_related_visit`).
    """
    if (
        not raw
        and not update_fields
        and not hasattr(instance, "metadata_create")
        and not instance._meta.label_lower.split(".")[1].startswith("historical")
    ):
        try:
            instance.metadata_update(entry_status=KEYED)
        except AttributeError as e:
            if "metadata_update" not in str(e):
                raise
        else:
            refresh_metadata_for_timepoint(instance, allow_create=True)


@receiver(post_delete, weak=False, dispatch_uid="metadata_reset_on_post_delete")
def metadata_reset_on_post_delete(sender, instance, using, **kwargs) -> None:
    """Deletes a single model instance used by UpdatesMetadataMixin.

    Not used by CrfMetadata and RequisitionMetadata.
    """
    try:
        instance.metadata_reset_on_delete()
    except AttributeError as e:
        if "metadata_reset_on_delete" not in str(e):
            raise
    else:
        refresh_metadata_for_timepoint(instance, allow_create=True)

    # deletes all for a visit used by CreatesMetadataMixin
    try:
        instance.metadata_delete_for_visit()
    except AttributeError as e:
        if "metadata_delete_for_visit" not in str(e):
            raise


@receiver(
    post_save,
    weak=False,
    dispatch_uid="metadata_update_previous_timepoints_for_singleton_on_post_save",
)
def metadata_update_previous_timepoints_for_singleton_on_post_save(
    sender, instance, raw, created, using, **kwargs
):
    if (
        not raw
        and not kwargs.get("update_fields")
        and isinstance(instance, (SingletonCrfModelMixin,))
    ):
        appointment = instance.related_visit.appointment.relative_previous_with_related_visit
        while appointment:
            if appointment.related_visit:
                refresh_metadata_for_timepoint(appointment.related_visit, allow_create=False)
            appointment = appointment.relative_previous_with_related_visit



from django.apps import apps as django_apps


def get_data_queries(subject_identifier=None, model=None, status=None, **kwargs):
    """Convenience func to return a data_query QuerySet."""
    model_cls = django_apps.get_model("edc_data_manager.dataquery")
    opts = {}
    if subject_identifier:
        opts.update(subject_identifier=subject_identifier)
    if model:
        opts.update(data_dictionaries__model=model)
    if status:
        opts.update(status=status)
    opts.update(**kwargs)
    return model_cls.objects.filter(**opts)



# Generated by Django 5.1.2 on 2024-11-11 14:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pylabels", "0009_remove_labelconfiguration_requires_allocated_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="labelconfiguration",
            name="requires_allocation",
            field=models.BooleanField(
                default=False, verbose_name="Configuration is for a subject label"
            ),
        ),
    ]



from django.apps import apps as django_apps
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.views.generic import TemplateView
from edc_dashboard.view_mixins import EdcViewMixin
from edc_navbar import NavbarViewMixin
from edc_protocol.view_mixins import EdcProtocolViewMixin

from ..constants import CENTRAL_LOCATION
from ..exceptions import StockTransferError
from ..models import Location, StockTransfer, StockTransferItem
from ..utils import transfer_stock_to_location


@method_decorator(login_required, name="dispatch")
class TransferStockView(EdcViewMixin, NavbarViewMixin, EdcProtocolViewMixin, TemplateView):
    """A view for transferring stock from central to a site.

    Creates a StockTransferItem instance per stock instance.

    See also: StockTransferConfirmationItem
    """

    model_pks: list[str] | None = None
    template_name: str = "edc_pharmacy/stock/transfer_stock.html"
    navbar_name = settings.APP_NAME
    navbar_selected_item = "pharmacy"

    def get_context_data(self, **kwargs):
        stock_transfer = StockTransfer.objects.get(pk=self.kwargs.get("stock_transfer"))
        transferred_count = StockTransferItem.objects.filter(
            stock_transfer=stock_transfer
        ).count()
        item_count = stock_transfer.item_count - transferred_count
        item_count = min(item_count, 12)
        kwargs.update(
            stock_transfer=stock_transfer,
            source_model_name=self.model_cls._meta.verbose_name_plural,
            source_changelist_url=self.source_changelist_url,
            from_locations=Location.objects.filter(site__isnull=True),
            to_locations=Location.objects.filter(site__isnull=False),
            item_count=list(range(1, item_count + 1)),
        )
        return super().get_context_data(**kwargs)

    @property
    def source_changelist_url(self):
        return reverse("edc_pharmacy_admin:edc_pharmacy_stocktransfer_changelist")

    @property
    def model_cls(self):
        return django_apps.get_model("edc_pharmacy.stocktransfer")

    def post(self, request, *args, **kwargs):  # noqa: ARG002
        stock_transfer = StockTransfer.objects.get(pk=self.kwargs.get("stock_transfer"))
        stock_codes = request.POST.getlist("codes")
        if stock_codes:
            transferred, dispensed_codes, skipped_codes, invalid_codes = [], [], [], []
            try:
                transferred, dispensed_codes, skipped_codes, invalid_codes = (
                    transfer_stock_to_location(stock_transfer, stock_codes, request=request)
                )
            except StockTransferError as e:
                messages.add_message(request, messages.ERROR, f"An error occured. {e}")

            if len(transferred) > 0:
                messages.add_message(
                    request,
                    messages.SUCCESS,
                    f"Successfully transferred {len(transferred)} stock items. ",
                )
            if skipped_codes:
                location = (
                    stock_transfer.to_location
                    if stock_transfer.from_location.name == CENTRAL_LOCATION
                    else stock_transfer.from_location
                )
                messages.add_message(
                    request,
                    messages.WARNING,
                    (
                        f"Skipped {len(skipped_codes)} "
                        f"stock item{'s' if len(skipped_codes) != 1 else ''}. "
                        f"Not for {location}. "
                        f"See {StockTransfer._meta.verbose_name} "
                        f"{stock_transfer.transfer_identifier}. "
                        f"Got {', '.join(skipped_codes)}"
                    ),
                )
            if dispensed_codes:
                messages.add_message(
                    request,
                    messages.ERROR,
                    f"Skipped {len(dispensed_codes)} stock item"
                    f"{'s' if len(dispensed_codes) != 1 else ''}. "
                    f"Already dispensed. Got {', '.join(dispensed_codes)}",
                )

            if invalid_codes:
                messages.add_message(
                    request,
                    messages.ERROR,
                    f"Skipped {len(invalid_codes)} stock item"
                    f"{'s' if len(invalid_codes) != 1 else ''}. "
                    f" are invalid. Got {', '.join(invalid_codes)}",
                )

        transferred_count = StockTransferItem.objects.filter(
            stock_transfer=stock_transfer
        ).count()
        if stock_transfer.item_count > transferred_count:
            url = reverse(
                "edc_pharmacy:transfer_stock_url",
                kwargs={
                    "stock_transfer": stock_transfer.id,
                },
            )
            return HttpResponseRedirect(url)
        url = f"{self.source_changelist_url}"
        return HttpResponseRedirect(url)
